#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Character and character-sequence data structures.
"""

import warnings
import copy
import math
import collections
from dendropy.utility.textprocessing import StringIO
from dendropy.utility import textprocessing
from dendropy.utility import error
from dendropy.utility import deprecate
from dendropy.utility import container
from dendropy.datamodel import charstatemodel
from dendropy.datamodel.charstatemodel import DNA_STATE_ALPHABET
from dendropy.datamodel.charstatemodel import RNA_STATE_ALPHABET
from dendropy.datamodel.charstatemodel import NUCLEOTIDE_STATE_ALPHABET
from dendropy.datamodel.charstatemodel import PROTEIN_STATE_ALPHABET
from dendropy.datamodel.charstatemodel import RESTRICTION_SITES_STATE_ALPHABET
from dendropy.datamodel.charstatemodel import INFINITE_SITES_STATE_ALPHABET
from dendropy.datamodel import basemodel
from dendropy.datamodel import taxonmodel
from dendropy import dataio

###############################################################################
## ContinuousCharElement

class ContinuousCharElement(
        basemodel.DataObject,
        basemodel.Annotable):
    def __init__(self, value, column_def=None, label=None):
        basemodel.DataObject.__init__(self,
                label=label)
        self.value = value
        self.column_def = column_def

###############################################################################
## CharacterType

class CharacterType(
        basemodel.DataObject,
        basemodel.Annotable):
    """
    A character format or type of a particular column: i.e., maps a particular
    set of character state definitions to a column in a character matrix.
    """

    def __init__(self,
            label=None,
            state_alphabet=None):
        basemodel.DataObject.__init__(self, label=label)
        self._state_alphabet = None
        self.state_alphabet = state_alphabet

    def _get_state_alphabet(self):
        """
        The |StateAlphabet| representing the state alphabet for this
        column: i.e., the collection of symbols and the state identities to
        which they map.
        """
        return self._state_alphabet
    def _set_state_alphabet(self, value):
        self._state_alphabet = value
    state_alphabet = property(_get_state_alphabet, _set_state_alphabet)

    def __copy__(self, memo=None):
        raise TypeError("Cannot directly copy {}".format(self.__class__.__name__))

    def taxon_namespace_scoped_copy(self, memo=None):
        raise TypeError("Cannot directly copy {}".format(self.__class__.__name__))

    def __deepcopy__(self, memo=None):
        return basemodel.Annotable.__deepcopy__(self, memo=memo)

###############################################################################
## CharacterDataSequence

class CharacterDataSequence(
        basemodel.Annotable,
        ):
    """
    A sequence of character values or values for a particular taxon or entry in
    a data matrix.

    Objects of this class can be (almost) treated as simple lists, where the
    elements are the values of characters (typically, real values in the case
    of continuous data, and special instances of |StateIdentity| objects in the
    case of discrete data.

    Character type data (represented by |CharacterType| instances) and metadata
    annotations (represented by |AnnotationSet| instances), if any, are
    maintained in a parallel list that need to be accessed separately using the
    index of the value to which the data correspond. So, for example, the
    |AnnotationSet| object containing the metadata annotations for the first
    value in a sequence, ``s[0]``, is available through
    ``s.annotations_at(0)``, while the character type information for that
    first element is available through ``s.character_type_at(0)`` and can be
    set through ``s.set_character_type_at(0, c)``.

    In most cases where metadata annotations and character type information are
    not needed, treating objects of this class as a simple list provides all
    the functionality needed. Where metadata annotations or character type
    information are required, all the standard list mutation methods (e.g.,
    ``CharacterDataSequence.insert``, ``CharacterDataSequence.append``,
    ``CharacterDataSequence.extend``) also take optional ``character_type``
    and ``character_annotations`` argument in addition to the primary
    ``character_value`` argument, thus allowing for setting of the value,
    character type, and annotation set simultaneously. While iteration over
    character values are available through the standard list iteration
    interface, the method ``CharacterDataSequence.cell_iter()`` provides for
    iterating over ``<character-value, character-type,
    character-annotation-set>`` triplets.

    """

    ###############################################################################
    ## Life-cycle

    def __init__(self,
            character_values=None,
            character_types=None,
            character_annotations=None):
        """
        Parameters
        ----------
        character_values : iterable of values
            A set of values for this sequence.
        """
        self._character_values = []
        self._character_types = []
        self._character_annotations = []
        if character_values:
            self.extend(
                    character_values=character_values,
                    character_types=character_types,
                    character_annotations=character_annotations)

    ###############################################################################
    ## Life-cycle

    # def values(self):
    #     return list(self._character_values)

    def values(self):
        """
        Returns list of values of this vector.

        Returns
        -------
        v : list
            List of values making up this vector.
        """
        return self._character_values

    def symbols_as_list(self):
        """
        Returns list of string representation of values of this vector.

        Returns
        -------
        v : list
            List of string representation of values making up this vector.
        """
        return list(str(cs) for cs in self._character_values)

    def symbols_as_string(self, sep=""):
        """
        Returns values of this vector as a single string, with individual value
        elements separated by ``sep``.

        Returns
        -------
        s : string
            String representation of values making up this vector.
        """
        return sep.join(str(cs) for cs in self._character_values)

    def __str__(self):
        return self.symbols_as_string()

    def append(self, character_value, character_type=None, character_annotations=None):
        """
        Adds a value to ``self``.

        Parameters
        ----------
        character_value : object
            Value to be stored.
        character_type : |CharacterType|
            Description of character value.
        character_annotations : |AnnotationSet|
            Metadata annotations associated with this character.
        """
        self._character_values.append(character_value)
        self._character_types.append(character_type)
        self._character_annotations.append(character_annotations)

    def extend(self, character_values, character_types=None, character_annotations=None):
        """
        Extends ``self`` with values.

        Parameters
        ----------
        character_values : iterable of objects
            Values to be stored.
        character_types : iterable of |CharacterType| objects
            Descriptions of character values.
        character_annotations : iterable |AnnotationSet| objects
            Metadata annotations associated with characters.
        """
        self._character_values.extend(character_values)
        if character_types is None:
            self._character_types.extend( [None] * len(character_values) )
        else:
            assert len(character_types) == len(character_values)
            self._character_types.extend(character_types)
        if character_annotations is None:
            self._character_annotations.extend( [None] * len(character_values) )
        else:
            assert len(character_annotations) == len(character_values)
            self._character_annotations.extend(character_annotations)

    def __len__(self):
        return len(self._character_values)

    def __getitem__(self, idx):
        return self._character_values[idx]

    def __setitem__(self, idx, value):
        self._character_values[idx] = value

    def __iter__(self):
        return self.__next__()

    def __next__(self):
        for v in self._character_values:
            yield v

    next = __next__ # Python 2 legacy support

    def cell_iter(self):
        """
        Iterate over triplets of character values and associated
        |CharacterType| and |AnnotationSet| instances.
        """
        for v, t, a in zip(self._character_values, self._character_types, self._character_annotations):
            yield v, t, a

    def __delitem__(self, idx):
        del self._character_values[idx]
        del self._character_types[idx]
        del self._character_annotations[idx]

    def set_at(self, idx, character_value, character_type=None, character_annotations=None):
        """
        Set value and associated character type and metadata annotations for
        element at ``idx``.

        Parameters
        ----------
        idx : integer
            Index of element to set.
        character_value : object
            Value to be stored.
        character_type : |CharacterType|
            Description of character value.
        character_annotations : |AnnotationSet|
            Metadata annotations associated with this character.
        """
        to_add = (idx+1) - len(self._character_values)
        while to_add > 0:
            self.append(None)
            to_add -= 1
        self._character_values[idx] = character_value
        self._character_types[idx] = character_type
        self._character_annotations[idx] = character_annotations

    def insert(self, idx, character_value, character_type=None, character_annotations=None):
        """
        Insert value and associated character type and metadata annotations for
        element at ``idx``.

        Parameters
        ----------
        idx : integer
            Index of element to set.
        character_value : object
            Value to be stored.
        character_type : |CharacterType|
            Description of character value.
        character_annotations : |AnnotationSet|
            Metadata annotations associated with this character.
        """
        self._character_values.insert(idx, character_value)
        self._character_types.insert(idx, character_type)
        self._character_annotations.insert(idx, character_annotations)

    def value_at(self, idx):
        """
        Return value of character at ``idx``.

        Parameters
        ----------
        idx : integer
            Index of element value to return.

        Returns
        -------
        c : object
            Value of character at index ``idx``.
        """
        return self._character_values[idx]

    def character_type_at(self, idx):
        """
        Return type of character at ``idx``.

        Parameters
        ----------
        idx : integer
            Index of element character type to return.

        Returns
        -------
        c : |CharacterType|
            |CharacterType| associated with character index ``idx``.
        """
        return self._character_types[idx]

    def annotations_at(self, idx):
        """
        Return metadata annotations of character at ``idx``.

        Parameters
        ----------
        idx : integer
            Index of element annotations to return.

        Returns
        -------
        c : |AnnotationSet|
            |AnnotationSet| representing metadata annotations of character at index ``idx``.
        """
        if self._character_annotations[idx] is None:
            self._character_annotations[idx] = basemodel.AnnotationSet(self._character_types[idx])
        return self._character_annotations[idx]

    def has_annotations_at(self, idx):
        """
        Return |True| if character at ``idx`` has metadata annotations.

        Parameters
        ----------
        idx : integer
            Index of element annotations to check.

        Returns
        -------
        b : bool
            |True| if character at ``idx`` has metadata annotations, |False|
            otherwise.
        """
        return not self._character_annotations[idx] is None

    def set_character_type_at(self, idx, character_type):
        """
        Set type of character at ``idx``.

        Parameters
        ----------
        idx : integer
            Index of element character type to set.
        """
        self._character_types[idx] = character_type

    def set_annotations_at(self, idx, annotations):
        """
        Set metadata annotations of character at ``idx``.

        Parameters
        ----------
        idx : integer
            Index of element annotations to set.
        """
        self._character_annotations[idx] = annotations

###############################################################################
## Subset of Character (Columns)

class CharacterSubset(
        basemodel.DataObject,
        basemodel.Annotable,
        ):
    """
    Tracks definition of a subset of characters.
    """

    def __init__(self, label=None, character_indices=None):
        """
        Parameters
        ----------
        label: str
            Name of this subset.
        character_indices: iterable of ``int``
            Iterable of 0-based (integer) indices of column positions that
            constitute this subset.

        """
        basemodel.DataObject.__init__(self, label=label)
        if character_indices is None:
            self.character_indices = set()
        else:
            self.character_indices = set(character_indices)

    def __len__(self):
        return len(self.character_indices)

    def __iter__(self):
        return iter(self.character_indices)

    def __deepcopy__(self, memo):
        return basemodel.Annotable.__deepcopy__(self, memo=memo)

###############################################################################
## CharacterMatrix

class CharacterMatrix(
        taxonmodel.TaxonNamespaceAssociated,
        basemodel.Annotable,
        basemodel.Deserializable,
        basemodel.NonMultiReadable,
        basemodel.Serializable,
        basemodel.DataObject):
    """
    A data structure that manages assocation of operational taxononomic unit
    concepts to sequences of character state identities or values.

    This is a base class that provides general functionality; derived classes
    specialize for particular data types. You will not be using the class
    directly, but rather one of the derived classes below, specialized for data
    types such as DNA, RNA, continuous, etc.

    This class and derived classes behave like a dictionary where the keys are
    |Taxon| objects and the values are `CharacterDataSequence` objects. Access
    to sequences based on taxon labels as well as indexes are also provided.
    Numerous methods are provided to manipulate and iterate over sequences.
    Character partitions can be managed through `CharacterSubset` objects,
    while management of detailed metadata on character types are available
    through |CharacterType| objects.

    Objects can be instantiated by reading data from external sources through
    the usual ``get_from_stream()``, ``get_from_path()``, or
    ``get_from_string()`` functions. In addition, a single matrix object can be
    instantiated from multiple matrices (``concatenate()``) or data sources
    (``concatenate_from_paths``).

    A range of methods also exist for importing data from another matrix object.
    These vary depending on how "new" and "existing" are treated.  A "new"
    sequence is a sequence in the other matrix associated with a |Taxon|
    object for which there is no sequence defined in the current matrix.  An
    "existing" sequence is a sequence in the other matrix associated with a
    |Taxon| object for which there *is* a sequence defined in the
    current matrix.

    +---------------------------------+---------------------------------------------+--------------------------------------------+
    |                                 | New Sequences: IGNORED                      | New Sequences: ADDED                       |
    +=================================+=============================================+============================================+
    | Existing Sequences: IGNORED     | [NO-OP]                                     | :meth:`CharacterMatrix.add_sequences()`    |
    +---------------------------------+---------------------------------------------+--------------------------------------------+
    | Existing Sequences: OVERWRITTEN | :meth:`CharacterMatrix.replace_sequences()` | :meth:`CharacterMatrix.update_sequences()` |
    +---------------------------------+---------------------------------------------+--------------------------------------------+
    | Existing Sequences: EXTENDED    | :meth:`CharacterMatrix.extend_sequences()`  | :meth:`CharacterMatrix.extend_matrix()`    |
    +---------------------------------+---------------------------------------------+--------------------------------------------+

    If character subsets have been defined, these subsets can be exported to independent matrices.

    """

    ###########################################################################
    ### Class Variables

    data_type = None
    character_sequence_type = CharacterDataSequence

    ###########################################################################
    ### Factory (Class) Methods

    def _parse_and_create_from_stream(cls,
            stream,
            schema,
            matrix_offset=0,
            **kwargs):
        taxon_namespace = taxonmodel.process_kwargs_dict_for_taxon_namespace(kwargs, None)
        if taxon_namespace is None:
            taxon_namespace = taxonmodel.TaxonNamespace(is_case_sensitive=kwargs.get("case_sensitive_taxon_labels", False))
        def tns_factory(label):
            if label is not None and taxon_namespace.label is None:
                taxon_namespace.label = label
            return taxon_namespace
        label = kwargs.pop("label", None)
        kwargs["data_type"] = cls.data_type
        reader = dataio.get_reader(schema, **kwargs)
        char_matrices = reader.read_char_matrices(
                stream=stream,
                taxon_namespace_factory=tns_factory,
                char_matrix_factory=new_char_matrix,
                state_alphabet_factory=charstatemodel.StateAlphabet,
                global_annotations_target=None)
        if len(char_matrices) == 0:
            raise ValueError("No character data in data source")
        char_matrix = char_matrices[matrix_offset]
        if char_matrix.data_type != cls.data_type:
            raise ValueError(
                "Data source (at offset {}) is of type '{}', "
                "but current CharacterMatrix is of type '{}'.".format(
                    matrix_offset,
                    char_matrix.data_type,
                    cls.data_type))
        return char_matrix
    _parse_and_create_from_stream = classmethod(_parse_and_create_from_stream)

    @classmethod
    def get(cls, **kwargs):
        r"""
        Instantiate and return a *new* character matrix object from a data source.

        **Mandatory Source-Specification Keyword Argument (Exactly One of the Following Required):**

            - **file** (*file*) -- File or file-like object of data opened for reading.
            - **path** (*str*) -- Path to file of data.
            - **url** (*str*) -- URL of data.
            - **data** (*str*) -- Data given directly.

        **Mandatory Schema-Specification Keyword Argument:**

            - **schema** (*str*) -- Identifier of format of data given by the
              "``file``", "``path``", "``data``", or "``url``" argument
              specified above: ":doc:`fasta </schemas/fasta>`", ":doc:`nexus
              </schemas/nexus>`", or ":doc:`nexml </schemas/nexml>`",
              ":doc:`phylip </schemas/phylip>`", etc.
              See "|Schemas|" for more details.

        **Optional General Keyword Arguments:**

            - **label** (*str*) -- Name or identifier to be assigned to the new
              object; if not given, will be assigned the one specified in the
              data source, or |None| otherwise.
            - **taxon_namespace** (|TaxonNamespace|) -- The |TaxonNamespace|
              instance to use to :doc:`manage the taxon names </primer/taxa>`.
              If not specified, a new one will be created.
            - **matrix_offset** (*int*) -- 0-based index of character block or
              matrix in source to be parsed. If not specified then the
              first matrix (offset = 0) is assumed.
            - **ignore_unrecognized_keyword_arguments** (*bool*) -- If |True|,
              then unsupported or unrecognized keyword arguments will not
              result in an error. Default is |False|: unsupported keyword
              arguments will result in an error.

        **Optional Schema-Specific Keyword Arguments:**

        These provide control over how the data is interpreted and
        processed, and supported argument names and values depend on
        the schema as specified by the value passed as the "``schema``"
        argument. See "|Schemas|" for more details.

        **Examples:**

        ::

            dna1 = dendropy.DnaCharacterMatrix.get(
                    file=open("pythonidae.fasta"),
                    schema="fasta")
            dna2 = dendropy.DnaCharacterMatrix.get(
                    url="http://purl.org/phylo/treebase/phylows/matrix/TB2:M2610?format=nexus",
                    schema="nexus")
            aa1 = dendropy.ProteinCharacterMatrix.get(
                    file=open("pythonidae.dat"),
                    schema="phylip")
            std1 = dendropy.StandardCharacterMatrix.get(
                    path="python_morph.nex",
                    schema="nexus")
            std2 = dendropy.StandardCharacterMatrix.get(
                    data=">t1\n01011\n\n>t2\n11100",
                    schema="fasta")

        """
        return cls._get_from(**kwargs)

    def concatenate(cls, char_matrices):
        """
        Creates and returns a single character matrix from multiple
        CharacterMatrix objects specified as a list, 'char_matrices'.
        All the CharacterMatrix objects in the list must be of the
        same type, and share the same TaxonNamespace reference. All taxa
        must be present in all alignments, all all alignments must
        be of the same length. Component parts will be recorded as
        character subsets.
        """
        taxon_namespace = char_matrices[0].taxon_namespace
        nseqs = len(char_matrices[0])
        concatenated_chars = cls(taxon_namespace=taxon_namespace)
        pos_start = 0
        for cidx, cm in enumerate(char_matrices):
            if cm.taxon_namespace is not taxon_namespace:
                raise ValueError("Different ``taxon_namespace`` references in matrices to be merged")
            if len(cm) != len(taxon_namespace):
                raise ValueError("Number of sequences not equal to the number of taxa")
            if len(cm) != nseqs:
                raise ValueError("Different number of sequences across alignments: %d (expecting %d based on first matrix)" % (len(cm), nseqs))
            v1 = len(cm[0])
            for t, s in cm.items():
                if len(s) != v1:
                    raise ValueError("Unequal length sequences in character matrix %d".format(cidx+1))
            concatenated_chars.extend_matrix(cm)
            if cm.label is None:
                new_label = "locus%03d" % cidx
            else:
                new_label = cm.label
            cs_label = new_label
            i = 2
            while cs_label in concatenated_chars.character_subsets:
                label = "%s_%03d" % (new_label, i)
                i += 1
            character_indices = range(pos_start, pos_start + cm.vector_size)
            pos_start += cm.vector_size
            concatenated_chars.new_character_subset(character_indices=character_indices,
                    label=cs_label)
        return concatenated_chars
    concatenate = classmethod(concatenate)

    def concatenate_from_streams(cls, streams, schema, **kwargs):
        """
        Read a character matrix from each file object given in ``streams``,
        assuming data format/schema ``schema``, and passing any keyword arguments
        down to the underlying specialized reader. Merge the character matrices
        and return the combined character matrix. Component parts will be
        recorded as character subsets.
        """
        taxon_namespace = taxonmodel.process_kwargs_dict_for_taxon_namespace(kwargs, None)
        if taxon_namespace is None:
            taxon_namespace = taxonmodel.TaxonNamespace(is_case_sensitive=kwargs.get("case_sensitive_taxon_labels", False))
        kwargs["taxon_namespace"] = taxon_namespace
        char_matrices = []
        for stream in streams:
            char_matrices.append(cls.get_from_stream(stream,
                schema=schema, **kwargs))
        return cls.concatenate(char_matrices)
    concatenate_from_streams = classmethod(concatenate_from_streams)

    def concatenate_from_paths(cls, paths, schema, **kwargs):
        """
        Read a character matrix from each file path given in ``paths``, assuming
        data format/schema ``schema``, and passing any keyword arguments down to
        the underlying specialized reader. Merge the and return the combined
        character matrix. Component parts will be recorded as character
        subsets.
        """
        try:
            streams = [open(path, "rU") for path in paths]
        except ValueError:
            streams = [open(path, "r") for path in paths]
        return cls.concatenate_from_streams(streams, schema, **kwargs)
    concatenate_from_paths = classmethod(concatenate_from_paths)

    def from_dict(cls,
            source_dict,
            char_matrix=None,
            case_sensitive_taxon_labels=False,
            **kwargs):
        r"""
        Populates character matrix from dictionary (or similar mapping type),
        creating |Taxon| objects and sequences as needed.

        Keys must be strings representing labels |Taxon| objects or
        |Taxon| objects directly. If key is specified as string, then it
        will be dereferenced to the first existing |Taxon| object in the
        current taxon namespace with the same label. If no such |Taxon|
        object can be found, then a new |Taxon| object is created and
        added to the current namespace. If a key is specified as a
        |Taxon| object, then this is used directly. If it is not in the
        current taxon namespace, it will be added.

        Values are the sequences (more generally, iterable of values).  If
        values are of type `CharacterDataSequence`, then they are added
        as-is.  Otherwise `CharacterDataSequence` instances are
        created for them. Values may be coerced into types compatible with
        particular matrices. The classmethod ``coerce_values()`` will be
        called for this.

        Examples
        --------

        The following creates a |DnaCharacterMatrix| instance with three
        sequences::

            d = {
                    "s1" : "TCCAA",
                    "s2" : "TGCAA",
                    "s3" : "TG-AA",
            }
            dna = DnaCharacterMatrix.from_dict(d)

        Three |Taxon| objects will be created, corresponding to the
        labels 's1', 's2', 's3'. Each associated string sequence will be
        converted to a `CharacterDataSequence`, with each symbol ("A", "C",
        etc.) being replaced by the DNA state represented by the symbol.

        Parameters
        ----------
        source_dict : dict or other mapping type
            Keys must be strings representing labels |Taxon| objects or
            |Taxon| objects directly. Values are sequences. See above
            for details.
        char_matrix : |CharacterMatrix|
            Instance of |CharacterMatrix| to populate with data. If not
            specified, a new one will be created using keyword arguments
            specified by ``kwargs``.
        case_sensitive_taxon_labels : boolean
            If |True|, matching of string labels specified as keys in ``d`` will
            be matched to |Taxon| objects in current taxon namespace
            with case being respected. If |False|, then case will be ignored.
        \*\*kwargs : keyword arguments, optional
            Keyword arguments to be passed to constructor of
            |CharacterMatrix| when creating new instance to populate, if
            no target instance is provided via ``char_matrix``.

        Returns
        -------
        char_matrix : |CharacterMatrix|
            |CharacterMatrix| populated by data from ``d``.
        """
        if char_matrix is None:
            char_matrix = cls(**kwargs)
        for key in source_dict:
            if textprocessing.is_str_type(key):
                taxon = char_matrix.taxon_namespace.require_taxon(key,
                        is_case_sensitive=case_sensitive_taxon_labels)
            else:
                taxon = key
                if taxon not in char_matrix.taxon_namespace:
                    char_matrix.taxon_namespace.add_taxon(taxon)
            s = char_matrix.coerce_values(source_dict[key])
            char_matrix[taxon] = s
        return char_matrix
    from_dict = classmethod(from_dict)

    ###########################################################################
    ### Lifecycle and Identity

    def __init__(self, *args, **kwargs):
        if len(args) > 1:
            # only allow 1 positional argument
            raise error.TooManyArgumentsError(func_name=self.__class__.__name__, max_args=1, args=args)
        elif len(args) == 1 and isinstance(args[0], CharacterMatrix):
            self._clone_from(args[0], kwargs)
        else:
            basemodel.DataObject.__init__(self, label=kwargs.pop("label", None))
            taxonmodel.TaxonNamespaceAssociated.__init__(self,
                    taxon_namespace=taxonmodel.process_kwargs_dict_for_taxon_namespace(kwargs, None))
            self._taxon_sequence_map = {}
            self.character_types = []
            self.comments = []
            self.character_subsets = container.OrderedCaselessDict()
            if len(args) == 1:
                # takes care of all possible initializations, including. e.g.,
                # tuples and so on
                d = collections.OrderedDict(args[0])
                self.__class__.from_dict(d, char_matrix=self)
        if kwargs:
            raise TypeError("Unrecognized or unsupported arguments: {}".format(kwargs))

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self is other

    def _clone_from(self, src, kwargs_dict):
        # super(Tree, self).__init__()
        memo = {}
        # memo[id(tree)] = self
        taxon_namespace = taxonmodel.process_kwargs_dict_for_taxon_namespace(kwargs_dict, src.taxon_namespace)
        memo[id(src.taxon_namespace)] = taxon_namespace
        if taxon_namespace is not src.taxon_namespace:
            for t1 in src.taxon_namespace:
                t2 = taxon_namespace.require_taxon(label=t1.label)
                memo[id(t1)] = t2
        else:
            for t1 in src.taxon_namespace:
                memo[id(t1)] = t1
        t = copy.deepcopy(src, memo)
        self.__dict__ = t.__dict__
        self.label = kwargs_dict.pop("label", src.label)
        return self

    def __copy__(self):
        other = self.__class__(label=self.label,
            taxon_namespace=self.taxon_namespace)
        for taxon in self._taxon_sequence_map:
            # other._taxon_sequence_map[taxon] = self.__class__.character_sequence_type(self._taxon_sequence_map[taxon])
            other._taxon_sequence_map[taxon] = self._taxon_sequence_map[taxon]
        memo = {}
        memo[id(self)] = other
        other.deep_copy_annotations_from(self, memo)
        return other

    def taxon_namespace_scoped_copy(self, memo=None):
        if memo is None:
            memo = {}
        # this populates ``memo`` with references to the
        # the TaxonNamespace and Taxon objects
        self.taxon_namespace.populate_memo_for_taxon_namespace_scoped_copy(memo)
        return self.__deepcopy__(memo=memo)

    def __deepcopy__(self, memo=None):
        return basemodel.Annotable.__deepcopy__(self, memo=memo)

    ###########################################################################
    ### Data I/O

    # def _parse_and_add_from_stream(self, stream, schema, **kwargs):
    #     """
    #     Populates objects of this type from ``schema``-formatted
    #     data in the file-like object source ``stream``, *replacing*
    #     all current data. If multiple character matrices are in the data
    #     source, a 0-based index of the character matrix to use can
    #     be specified using the ``matrix_offset`` keyword (defaults to 0, i.e., first
    #     character matrix).
    #     """
    #     warnings.warn("Repopulating a CharacterMatrix is now deprecated. Instantiate a new instance from the source instead.",
    #             DeprecationWarning)
    #     m = self.__class__._parse_and_create_from_stream(stream=stream,
    #             schema=schema,
    #             **kwargs)
    #     return self.clone_from(m)

    def _format_and_write_to_stream(self, stream, schema, **kwargs):
        r"""
        Writes out ``self`` in ``schema`` format to a destination given by
        file-like object ``stream``.

        Parameters
        ----------
        stream : file or file-like object
            Destination for data.
        schema : string
            Must be a recognized character file schema, such as "nexus",
            "phylip", etc, for which a specialized writer is available. If this
            is not implemented for the schema specified, then a
            UnsupportedSchemaError is raised.

        \*\*kwargs : keyword arguments, optional
            Keyword arguments will be passed directly to the writer for the
            specified schema. See documentation for details on keyword
            arguments supported by writers of various schemas.

        """
        writer = dataio.get_writer(schema, **kwargs)
        writer.write_char_matrices([self],
                stream)

    ###########################################################################
    ### Taxon Management

    def reconstruct_taxon_namespace(self,
            unify_taxa_by_label=True,
            taxon_mapping_memo=None):
        """
        See `TaxonNamespaceAssociated.reconstruct_taxon_namespace`.
        """
        if taxon_mapping_memo is None:
            taxon_mapping_memo = {}
        original_taxa = list(self._taxon_sequence_map.keys())
        for original_taxon in original_taxa:
            if unify_taxa_by_label or original_taxon not in self.taxon_namespace:
                t = taxon_mapping_memo.get(original_taxon, None)
                if t is None:
                    # taxon to use not given and
                    # we have not yet created a counterpart
                    if unify_taxa_by_label:
                        # this will force usage of any taxon with
                        # a label that matches the current taxon
                        t = self.taxon_namespace.require_taxon(label=original_taxon.label)
                    else:
                        # this will unconditionally create a new taxon
                        t = self.taxon_namespace.new_taxon(label=original_taxon.label)
                    taxon_mapping_memo[original_taxon] = t
                else:
                    # taxon to use is given by mapping
                    self.taxon_namespace.add_taxon(t)
                if t in self._taxon_sequence_map:
                    raise error.TaxonNamespaceReconstructionError("Multiple sequences for taxon with label '{}'".format(t.label))
                self._taxon_sequence_map[t] = self._taxon_sequence_map[original_taxon]
                del self._taxon_sequence_map[original_taxon]

    def poll_taxa(self, taxa=None):
        """
        Returns a set populated with all of |Taxon| instances associated
        with ``self``.

        Parameters
        ----------
        taxa : set()
            Set to populate. If not specified, a new one will be created.

        Returns
        -------
        taxa : set[|Taxon|]
            Set of taxa associated with ``self``.
        """
        if taxa is None:
            taxa = set()
        for taxon in self._taxon_sequence_map:
            taxa.add(taxon)
        return taxa

    def update_taxon_namespace(self):
        """
        All |Taxon| objects in ``self`` that are not in
        ``self.taxon_namespace`` will be added.
        """
        assert self.taxon_namespace is not None
        for taxon in self._taxon_sequence_map:
            if taxon not in self.taxon_namespace:
                self.taxon_namespace.add_taxon(taxon)

    def reindex_subcomponent_taxa(self):
        """
        Synchronizes |Taxon| objects of map to ``taxon_namespace`` of self.
        """
        raise NotImplementedError("'reindex_subcomponent_taxa()' is no longer supported; use '{}.reconstruct_taxon_namespace()' instead".format(self.__class__.__name__))

    ###########################################################################
    ### Sequence CRUD

    def _resolve_key(self, key):
        """
        Resolves map access key into |Taxon| instance.

        If ``key`` is integer, assumed to be taxon index.
        If ``key`` string, assumed to be taxon label.
        Otherwise, assumed to be |Taxon| instance directly.
        """
        if isinstance(key, int):
            if abs(key) < len(self.taxon_namespace):
                taxon = self.taxon_namespace[key]
            else:
                raise IndexError(key)
        elif textprocessing.is_str_type(key):
            taxon = self.taxon_namespace.get_taxon(label=key)
            if taxon is None:
                raise KeyError(key)
        else:
            taxon = key
        return taxon

    def new_sequence(self, taxon, values=None):
        """
        Creates a new `CharacterDataSequence` associated with |Taxon|
        ``taxon``, and populates it with values in ``values``.

        Parameters
        ----------
        taxon : |Taxon|
            |Taxon| instance with which this sequence is associated.
        values : iterable or |None|
            An initial set of values with which to populate the new character
            sequence.

        Returns
        -------
        s : `CharacterDataSequence`
            A new `CharacterDataSequence` associated with |Taxon|
            ``taxon``.
        """
        if taxon in self._taxon_sequence_map:
            raise ValueError("Character values vector for taxon {} already exists".format(repr(taxon)))
        if taxon not in self.taxon_namespace:
            raise ValueError("Taxon {} is not in object taxon namespace".format(repr(taxon)))
        cv = self.__class__.character_sequence_type(values)
        self._taxon_sequence_map[taxon] = cv
        return cv

    def __getitem__(self, key):
        """
        Retrieves sequence for ``key``, which can be a index or a label of a
        |Taxon| instance in the current taxon namespace, or a
        |Taxon| instance directly.

        If no sequence is currently associated with specified |Taxon|, a
        new one will be created. Note that the |Taxon| object must have
        already been defined in the curent taxon namespace.

        Parameters
        ----------
        key : integer, string, or |Taxon|
            If an integer, assumed to be an index of a |Taxon| object in
            the current |TaxonNamespace| object of ``self.taxon_namespace``.
            If a string, assumed to be a label of a |Taxon| object in
            the current |TaxonNamespace| object of ``self.taxon_namespace``.
            Otherwise, assumed to be |Taxon| instance directly. In all
            cases, the |Taxon| object must be (already) defined in the
            current taxon namespace.

        Returns
        -------
        s : `CharacterDataSequence`
            A sequence associated with the |Taxon| instance referenced
            by ``key``.
        """
        taxon = self._resolve_key(key)
        try:
            return self._taxon_sequence_map[taxon]
        except KeyError:
            return self.new_sequence(taxon)

    def __setitem__(self, key, values):
        """
        Assigns sequence ``values`` to taxon specified by ``key``, which can be a
        index or a label of a |Taxon| instance in the current taxon
        namespace, or a |Taxon| instance directly.

        If no sequence is currently associated with specified |Taxon|, a
        new one will be created.  Note that the |Taxon| object must have
        already been defined in the curent taxon namespace.

        Parameters
        ----------
        key : integer, string, or |Taxon|
            If an integer, assumed to be an index of a |Taxon| object in
            the current |TaxonNamespace| object of ``self.taxon_namespace``.
            If a string, assumed to be a label of a |Taxon| object in
            the current |TaxonNamespace| object of ``self.taxon_namespace``.
            Otherwise, assumed to be |Taxon| instance directly. In all
            cases, the |Taxon| object must be (already) defined in the
            current taxon namespace.

        """
        taxon = self._resolve_key(key)
        if taxon not in self.taxon_namespace:
            raise ValueError(repr(key))
        if not isinstance(values, self.__class__.character_sequence_type):
            values = self.__class__.character_sequence_type(values)
        self._taxon_sequence_map[taxon] = values

    def __contains__(self, key):
        """
        Returns |True| if a sequence associated with ``key`` is in ``self``, or
        |False| otherwise.

        Parameters
        ----------
        key : integer, string, or |Taxon|
            If an integer, assumed to be an index of a |Taxon| object in
            the current |TaxonNamespace| object of ``self.taxon_namespace``.
            If a string, assumed to be a label of a |Taxon| object in
            the current |TaxonNamespace| object of ``self.taxon_namespace``.
            Otherwise, assumed to be |Taxon| instance directly. In all
            cases, the |Taxon| object must be (already) defined in the
            current taxon namespace.

        Returns
        -------
        b : boolean
            |True| if ``key`` is in ``self``; |False| otherwise.
        """
        return self._taxon_sequence_map.__contains__(key)

    def __delitem__(self, key):
        """
        Removes sequence for ``key``, which can be a index or a label of a
        |Taxon| instance in the current taxon namespace, or a
        |Taxon| instance directly.

        Parameters
        ----------
        key : integer, string, or |Taxon|
            If an integer, assumed to be an index of a |Taxon| object in
            the current |TaxonNamespace| object of ``self.taxon_namespace``.
            If a string, assumed to be a label of a |Taxon| object in
            the current |TaxonNamespace| object of ``self.taxon_namespace``.
            Otherwise, assumed to be |Taxon| instance directly. In all
            cases, the |Taxon| object must be (already) defined in the
            current taxon namespace.

        """
        return self._taxon_sequence_map.__delitem__(key)

    def clear(self):
        """
        Removes all sequences from matrix.
        """
        self._taxon_sequence_map.clear()

    def sequences(self):
        """
        List of all sequences in self.

        Returns
        -------
        s : list of `CharacterDataSequence` objects in self

        """
        s = [self[taxon] for taxon in self]
        return s

    def vectors(self):
        deprecate.dendropy_deprecation_warning(
                message="Deprecated since DendroPy 4: 'vectors()' will no longer be supported in future releases; use 'sequences()' instead")
        return self.sequences()

    ###########################################################################
    ### Symbol/alphabet management

    def coerce_values(self, values):
        """
        Converts elements of ``values`` to type of matrix.

        This method is called by :meth:`CharacterMatrix.from_dict` to create
        sequences from iterables of values.  This method should be overridden
        by derived classes to ensure that ``values`` consists of types compatible
        with the particular type of matrix. For example, a CharacterMatrix type
        with a fixed state alphabet (such as |DnaCharacterMatrix|) would
        dereference the string elements of ``values`` to return a list of
        |StateIdentity| objects corresponding to the symbols represented
        by the strings.  If there is no value-type conversion done, then
        ``values`` should be returned as-is. If no value-type conversion is
        possible (e.g., when the type of a value is dependent on positionaly
        information), then a TypeError should be raised.

        Parameters
        ----------
        values : iterable
            Iterable of values to be converted.

        Returns
        -------
        v : list of values.
        """
        return values

    ###########################################################################
    ### Sequence Access Iteration

    def __iter__(self):
        "Returns an iterator over character map's ordered keys."
        for t in self.taxon_namespace:
            if t in self._taxon_sequence_map:
                yield t

    def __contains__(self, key):
        if isinstance(key, str):
            for t in self._taxon_sequence_map:
                if t.label == key:
                    return True
            return False
        return key in self._taxon_sequence_map

    def values(self):
        """
        Iterates values (i.e. sequences) in this matrix.
        """
        for t in self:
            yield self[t]

    # def iterkeys(self):
    #     "Dictionary interface implementation for direct access to character map."
    #     for t in self.taxon_namespace:
    #         if t in self._taxon_sequence_map:
    #             yield t

    # def itervalues(self):
    #     "Dictionary interface implementation for direct access to character map."
    #     for t in self.taxon_namespace:
    #         if t in self._taxon_sequence_map:
    #             yield self._taxon_sequence_map[t]

    def items(self):
        "Returns character map key, value pairs in key-order."
        for t in self.taxon_namespace:
            if t in self._taxon_sequence_map:
                yield t, self._taxon_sequence_map[t]

    # def values(self):
    #     "Returns list of values."
    #     return [self._taxon_sequence_map[t] for t in self.taxon_namespace if t in self._taxon_seq_map]

    # def pop(self, key, alt_val=None):
    #     "a.pop(k[, x]):  a[k] if k in a, else x (and remove k)"
    #     return self._taxon_sequence_map.pop(key, alt_val)

    # def popitem(self):
    #     "a.popitem()  remove and last (key, value) pair"
    #     return self._taxon_sequence_map.popitem()

    # def keys(self):
    #     "Returns a copy of the ordered list of character map keys."
    #     return list(self._taxon_sequence_map.keys())

    ###########################################################################
    ### Metrics

    def __len__(self):
        """
        Number of sequences in matrix.

        Returns
        -------
        n : Number of sequences in matrix.
        """
        return len(self._taxon_sequence_map)

    def _get_sequence_size(self):
        """
        Number of characters in *first* sequence in matrix.

        Returns
        -------
        n : integer
            Number of sequences in matrix.
        """
        if len(self):
            # yuck, but len(self.values())
            # means we have to create and populate a list ...
            return len(self[next(iter(self._taxon_sequence_map))])
        else:
            return 0
    sequence_size = property(_get_sequence_size, None, None)
    vector_size = property(_get_sequence_size, None, None) # legacy

    def _get_max_sequence_size(self):
        """
        Maximum number of characters across all sequences in matrix.

        Returns
        -------
        n : integer
            Maximum number of characters across all sequences in matrix.
        """
        max_len = 0
        for k in self:
            if len(self[k]) > max_len:
                max_len  = len(self._taxon_sequence_map[k])
        return max_len
    max_sequence_size = property(_get_max_sequence_size, None, None)

    ###########################################################################
    ### Mass/Bulk Operations

    def fill(self, value, size=None, append=True):
        """
        Pads out all sequences in ``self`` by adding ``value`` to each sequence
        until its length is ``size`` long or equal to the length of the longest
        sequence if ``size`` is not specified.

        Parameters
        ----------
        value : object
            A valid value (e.g., a numeric value for continuous characters, or
            a |StateIdentity| for discrete character).
        size : integer or None
            The size (length) up to which the sequences will be padded. If |None|, then
            the maximum (longest) sequence size will be used.
        append : boolean
            If |True| (default), then new values will be added to the end of
            each sequence. If |False|, then new values will be inserted to the
            front of each sequence.
        """
        if size is None:
            size = self.max_sequence_size
        for k in self:
            v = self[k]
            while len(v) < size:
                if append:
                    v.append(value)
                else:
                    v.insert(0, value)
        return size

    def fill_taxa(self):
        """
        Adds a new (empty) sequence for each |Taxon| instance in
        current taxon namespace that does not have a sequence.
        """
        for taxon in self.taxon_namespace:
            if taxon not in self:
                self[taxon] = CharacterDataSequence()

    def pack(self, value=None, size=None, append=True):
        """
        Adds missing sequences for all |Taxon| instances in current
        namespace, and then pads out all sequences in ``self`` by adding ``value``
        to each sequence until its length is ``size`` long or equal to the length
        of the longest sequence if ``size`` is not specified. A combination of
        :meth:`CharacterMatrix.fill_taxa()` and
        :meth:`CharacterMatrix.fill()`.

        Parameters
        ----------
        value : object
            A valid value (e.g., a numeric value for continuous characters, or
            a |StateIdentity| for discrete character).
        size : integer or None
            The size (length) up to which the sequences will be padded. If |None|, then
            the maximum (longest) sequence size will be used.
        append : boolean
            If |True| (default), then new values will be added to the end of
            each sequence. If |False|, then new values will be inserted to the
            front of each sequence.
        """
        self.fill_taxa()
        self.fill(value=value, size=size, append=append)

    def add_sequences(self, other_matrix):
        """
        Adds sequences for |Taxon| objects that are in ``other_matrix`` but not in
        ``self``.

        Parameters
        ----------
        other_matrix : |CharacterMatrix|
            Matrix from which to add sequences.

        Notes
        -----
            1. ``other_matrix`` must be of same type as ``self``.
            2. ``other_matrix`` must have the same |TaxonNamespace| as ``self``.
            3. Each sequence associated with a |Taxon| reference in ``other_matrix``
               but not in ``self`` will be added to ``self`` as a shallow-copy.
            4. All other sequences will be ignored.

        """
        if other_matrix.taxon_namespace is not self.taxon_namespace:
            raise error.TaxonNamespaceIdentityError(self, other_matrix)
        for taxon in other_matrix._taxon_sequence_map:
            if taxon not in self._taxon_sequence_map:
                self._taxon_sequence_map[taxon] = self.__class__.character_sequence_type(other_matrix._taxon_sequence_map[taxon])

    def replace_sequences(self, other_matrix):
        """
        Replaces sequences for |Taxon| objects shared between ``self`` and
        ``other_matrix``.

        Parameters
        ----------
        other_matrix : |CharacterMatrix|
            Matrix from which to replace sequences.

        Notes
        -----
            1. ``other_matrix`` must be of same type as ``self``.
            2. ``other_matrix`` must have the same |TaxonNamespace| as ``self``.
            3. Each sequence in ``self`` associated with a |Taxon| that is
               also represented in ``other_matrix`` will be replaced with a
               shallow-copy of the corresponding sequence from ``other_matrix``.
            4. All other sequences will be ignored.
        """
        if other_matrix.taxon_namespace is not self.taxon_namespace:
            raise error.TaxonNamespaceIdentityError(self, other_matrix)
        for taxon in other_matrix._taxon_sequence_map:
            if taxon in self._taxon_sequence_map:
                self._taxon_sequence_map[taxon] = self.__class__.character_sequence_type(other_matrix._taxon_sequence_map[taxon])

    def update_sequences(self, other_matrix):
        """
        Replaces sequences for |Taxon| objects shared between ``self`` and
        ``other_matrix`` and adds sequences for |Taxon| objects that are
        in ``other_matrix`` but not in ``self``.

        Parameters
        ----------
        other_matrix : |CharacterMatrix|
            Matrix from which to update sequences.

        Notes
        -----
            1. ``other_matrix`` must be of same type as ``self``.
            2. ``other_matrix`` must have the same |TaxonNamespace| as ``self``.
            3. Each sequence associated with a |Taxon| reference in ``other_matrix``
               but not in ``self`` will be added to ``self``.
            4. Each sequence in ``self`` associated with a |Taxon| that is
               also represented in ``other_matrix`` will be replaced with a
               shallow-copy of the corresponding sequence from ``other_matrix``.
        """
        if other_matrix.taxon_namespace is not self.taxon_namespace:
            raise error.TaxonNamespaceIdentityError(self, other_matrix)
        for taxon in other_matrix._taxon_sequence_map:
            self._taxon_sequence_map[taxon] = self.__class__.character_sequence_type(other_matrix._taxon_sequence_map[taxon])

    def extend_sequences(self, other_matrix, is_add_new_sequences=False):
        """
        Extends sequences in ``self`` with characters associated with
        corresponding |Taxon| objects in ``other_matrix``.

        Parameters
        ----------
        other_matrix : |CharacterMatrix|
            Matrix from which to extend sequences.

        Notes
        -----
            1. ``other_matrix`` must be of same type as ``self``.
            2. ``other_matrix`` must have the same |TaxonNamespace| as ``self``.
            3. Each sequence associated with a |Taxon| reference in
               ``other_matrix`` that is also in ``self`` will be appended to the
               sequence currently associated with that |Taxon| reference
               in ``self``.
            4. All other sequences will be ignored.
        """
        if other_matrix.taxon_namespace is not self.taxon_namespace:
            raise error.TaxonNamespaceIdentityError(self, other_matrix)
        for taxon in other_matrix._taxon_sequence_map:
            if taxon not in self._taxon_sequence_map:
                if not is_add_new_sequences:
                    continue
                self._taxon_sequence_map[taxon] = self.__class__.character_sequence_type(other_matrix._taxon_sequence_map[taxon])
            else:
                self._taxon_sequence_map[taxon].extend(other_matrix._taxon_sequence_map[taxon])

    def extend_matrix(self, other_matrix):
        """
        Extends sequences in ``self`` with characters associated with
        corresponding |Taxon| objects in ``other_matrix`` and adds
        sequences for |Taxon| objects that are in ``other_matrix`` but not
        in ``self``.

        Parameters
        ----------
        other_matrix : |CharacterMatrix|
            Matrix from which to extend.

        Notes
        -----
            1. ``other_matrix`` must be of same type as ``self``.
            2. ``other_matrix`` must have the same |TaxonNamespace| as ``self``.
            3. Each sequence associated with a |Taxon| reference in ``other_matrix``
               that is also in ``self`` will be appending
               to the sequence currently associated with that |Taxon|
               reference in ``self``.
            4. Each sequence associated with a |Taxon| reference in
               ``other_matrix`` that is also in ``self`` will replace the sequence
               currently associated with that |Taxon| reference in ``self``.
        """
        if other_matrix.taxon_namespace is not self.taxon_namespace:
            raise error.TaxonNamespaceIdentityError(self, other_matrix)
        for taxon in other_matrix._taxon_sequence_map:
            if taxon in self._taxon_sequence_map:
                self._taxon_sequence_map[taxon].extend(other_matrix._taxon_sequence_map[taxon])
            else:
                self._taxon_sequence_map[taxon]= self.__class__.character_sequence_type(other_matrix._taxon_sequence_map[taxon])

    def remove_sequences(self, taxa):
        """
        Removes sequences associated with |Taxon| instances specified in
        ``taxa``. A KeyError is raised if a |Taxon| instance is
        specified for which there is no associated sequences.

        Parameters
        ----------
        taxa : iterable[|Taxon|]
            List or some other iterable of |Taxon| instances.
        """
        for taxon in taxa:
            del self._taxon_sequence_map[taxon]

    def discard_sequences(self, taxa):
        """
        Removes sequences associated with |Taxon| instances specified in
        ``taxa`` if they exist.

        Parameters
        ----------
        taxa : iterable[|Taxon|]
            List or some other iterable of |Taxon| instances.
        """
        for taxon in taxa:
            try:
                del self._taxon_sequence_map[taxon]
            except KeyError:
                pass

    def keep_sequences(self, taxa):
        """
        Discards all sequences *not* associated with any of the |Taxon| instances.

        Parameters
        ----------
        taxa : iterable[|Taxon|]
            List or some other iterable of |Taxon| instances.
        """
        to_keep = set(taxa)
        for taxon in tuple(self._taxon_sequence_map.keys()):
            if taxon not in to_keep:
                del self._taxon_sequence_map[taxon]

    # def extend_characters(self, other_matrix):
    #     """
    #     DEPRECATED
    #     Extends this matrix by adding characters from sequences of taxa
    #     in given matrix to sequences of taxa with correspond labels in
    #     this one. Taxa in the second matrix that do not exist in the
    #     current one are ignored.
    #     """
    #     self._taxon_sequence_map.extend_characters(other_matrix.taxon_seq_map)

    # def extend_map(self,
    #                   other_map,
    #                   overwrite_existing=False,
    #                   extend_existing=False):
    #     """
    #     DEPRECATED
    #     Extends this matrix by adding taxa and characters from the given
    #     map to this one.  If ``overwrite_existing`` is True and a taxon
    #     in the other map is already present in the current one, then
    #     the sequence associated with the taxon in the second map
    #     replaces the sequence in the current one. If ``extend_existing``
    #     is True and a taxon in the other matrix is already present in
    #     the current one, then the squence map with the taxon in
    #     the second map will be added to the sequence in the current
    #     one. If both are True, then an exception is raised. If neither
    #     are True,  and a taxon in the other map is already present in
    #     the current one, then the sequence is ignored.
    #     """
    #     self._taxon_sequence_map.extend(other_map,
    #         overwrite_existing=overwrite_existing,
    #         extend_existing=extend_existing)
    #     self.update_taxon_namespace()

    # def extend(self,
    #            other_matrix,
    #            overwrite_existing=False,
    #            extend_existing=False):
    #     """
    #     Extends this matrix by adding taxa and characters from the given
    #     matrix to this one.  If ``overwrite_existing`` is True and a taxon
    #     in the other matrix is already present in the current one, then
    #     the sequence associated with the taxon in the second matrix
    #     replaces the sequence in the current one. If ``extend_existing``
    #     is True and a taxon in the other matrix is already present in
    #     the current one, then the sequence associated with the taxon in
    #     the second matrix will be added to the sequence in the current
    #     one. If both are True, then an exception is raised. If neither
    #     are True, and a taxon in the other matrix is already present in
    #     the current one, then the sequence is ignored.
    #     """
    #     self._taxon_sequence_map.extend(other_matrix.taxon_seq_map,
    #         overwrite_existing=overwrite_existing,
    #         extend_existing=extend_existing)
    #     self.update_taxon_namespace()

    ###########################################################################
    ### Character Subset Management

    def add_character_subset(self, char_subset):
        """
        Adds a CharacterSubset object. Raises an error if one already exists
        with the same label.
        """
        label = char_subset.label
        if label in self.character_subsets:
            raise ValueError("Character subset '%s' already defined" % label)
        self.character_subsets[label] = char_subset
        return self.character_subsets[label]

    def new_character_subset(self, label, character_indices):
        """
        Defines a set of character (columns) that make up a character set.
        Raises an error if one already exists with the same label. Column
        indices are 0-based.
        """
        cs = CharacterSubset(character_indices=character_indices, label=label)
        return self.add_character_subset(cs)

    ###########################################################################
    ### CharacterType Management

    def new_character_type(self, *args, **kwargs):
        return CharacterType(*args, **kwargs)

    ###########################################################################
    ### Export

    def export_character_subset(self, character_subset):
        """
        Returns a new CharacterMatrix (of the same type) consisting only
        of columns given by the CharacterSubset, ``character_subset``.
        Note that this new matrix will still reference the same taxon set.
        """
        if textprocessing.is_str_type(character_subset):
            if character_subset not in self.character_subsets:
                raise KeyError(character_subset)
            else:
                character_subset = self.character_subsets[character_subset]
        return self.export_character_indices(character_subset.character_indices)

    def export_character_indices(self, indices):
        """
        Returns a new CharacterMatrix (of the same type) consisting only
        of columns given by the 0-based indices in ``indices``.
        Note that this new matrix will still reference the same taxon set.
        """
        clone = self.__class__(self)
        # clear out character subsets; otherwise all indices will have to be
        # recalculated, which will require some careful and perhaps arbitrary
        # handling of corner cases
        clone.character_subsets = container.OrderedCaselessDict()
        indices = set(indices)
        for vec in clone.values():
            for cell_idx in range(len(vec)-1, -1, -1):
                if cell_idx not in indices:
                    del(vec[cell_idx])
        return clone

    ###########################################################################
    ### Representation

    def description(self, depth=1, indent=0, itemize="", output=None):
        """
        Returns description of object, up to level ``depth``.
        """
        if depth is None or depth < 0:
            return
        output_strio = StringIO()
        label = " (%s: '%s')" % (id(self), self.label)
        output_strio.write('%s%s%s object at %s%s'
                % (indent*' ',
                   itemize,
                   self.__class__.__name__,
                   hex(id(self)),
                   label))
        if depth >= 1:
            output_strio.write(':  %d Sequences' % len(self))
            if depth >= 2:
                if self.taxon_namespace is not None:
                    tlead = "\n%s[Taxon Set]\n" % (" " * (indent+4))
                    output_strio.write(tlead)
                    self.taxon_namespace.description(depth=depth-1, indent=indent+8, itemize="", output=output_strio)
                tlead = "\n%s[Characters]\n" % (" " * (indent+4))
                output_strio.write(tlead)
                indent += 8
                maxlabel = max([len(str(t.label)) for t in self.taxon_namespace])
                for i, t in enumerate(self.taxon_namespace):
                    output_strio.write('%s%s%s : %s characters\n' \
                        % (" " * indent,
                           "[%d] " % i,
                           str(t.label),
                           len(self._taxon_sequence_map[t])))

        s = output_strio.getvalue()
        if output is not None:
            output.write(s)
        return s

    ###########################################################################
    ### Legacy

    def _get_taxon_seq_map(self):
        warnings.warn("All methods and features of 'CharacterMatrix.taxon_seq_map' have been integrated directly into 'CharacterMatrix', or otherwise replaced entirely",
                stacklevel=2)
        return self
    taxon_seq_map = property(_get_taxon_seq_map)

###############################################################################
## Specialized Matrices

### Continuous Characters ##################################################

class ContinuousCharacterDataSequence(CharacterDataSequence):
    """
    A sequence of continuous character values for a particular taxon or entry
    in a data matrix. Specializes `CharacterDataSequence` by assuming all
    values are primitive numerics (i.e., either floats or integers) when
    copying or representing self.
    """

    def symbols_as_list(self):
        """
        Returns list of string representation of values of this vector.

        Returns
        -------
        v : list
            List of string representation of values making up this vector.
        """
        return [str(v) for v in self]

    def symbols_as_string(self, sep=" "):
        # different default
        return CharacterDataSequence.symbols_as_string(self, sep=sep)

class ContinuousCharacterMatrix(CharacterMatrix):
    """
    Specializes |CharacterMatrix| for continuous data.

    Sequences stored using |ContinuousCharacterDataSequence|, with values of
    elements assumed to be ``float`` .
    """

    character_sequence_type = ContinuousCharacterDataSequence
    data_type = "continuous"

    def __init__(self, *args, **kwargs):
        CharacterMatrix.__init__(self, *args, **kwargs)

### Discrete Characters ##################################################

class DiscreteCharacterDataSequence(CharacterDataSequence):
    pass

class DiscreteCharacterMatrix(CharacterMatrix):

    character_sequence_type = DiscreteCharacterDataSequence

    data_type = "discrete"

    def __init__(self, *args, **kwargs):
        CharacterMatrix.__init__(self, *args, **kwargs)
        self.state_alphabets = []
        self._default_state_alphabet = None

    def _get_default_state_alphabet(self):
        if self._default_state_alphabet is not None:
            return self._default_state_alphabet
        elif len(self.state_alphabets) == 1:
            return self.state_alphabets[0]
        elif len(self.state_alphabets) > 1:
            raise TypeError("Multiple state alphabets defined for this matrix with no default specified")
        elif len(self.state_alphabets) == 0:
            raise TypeError("No state alphabets defined for this matrix")
        return None
    def _set_default_state_alphabet(self, s):
        if s not in self.state_alphabets:
            self.state_alphabets.append(s)
        self._default_state_alphabet = s
    default_state_alphabet = property(_get_default_state_alphabet, _set_default_state_alphabet)

    def append_taxon_sequence(self, taxon, state_symbols):
        if taxon not in self:
            self[taxon] = CharacterDataSequence()
        for value in state_symbols:
            if textprocessing.is_str_type(value):
                symbol = value
            else:
                symbol = str(value)
            self[taxon].append(self.default_symbol_state_map[symbol])

    def remap_to_state_alphabet_by_symbol(self,
            state_alphabet,
            purge_other_state_alphabets=True):
        """
        All entities with any reference to a state alphabet will be have the
        reference reassigned to state alphabet ``sa``, and all entities with
        any reference to a state alphabet element will be have the reference
        reassigned to any state alphabet element in ``sa`` that has the same
        symbol. Raises KeyError if no matching symbol can be found.
        """
        for vi, vec in enumerate(self._taxon_sequence_map.values()):
            for ci, cell in enumerate(vec):
                vec[ci] = state_alphabet[cell.symbol]
        for ct in self.character_types:
            if ct is not None:
                ct.state_alphabet = state_alphabet
        if purge_other_state_alphabets:
            self.default_state_alphabet = state_alphabet

    def remap_to_default_state_alphabet_by_symbol(self,
            purge_other_state_alphabets=True):
        """
        All entities with any reference to a state alphabet will be have the
        reference reassigned to the default state alphabet, and all entities
        with any reference to a state alphabet element will be have the
        reference reassigned to any state alphabet element in the default
        state alphabet that has the same symbol. Raises ValueError if no
        matching symbol can be found.
        """
        self.remap_to_state_alphabet_by_symbol(
                state_alphabet=self.default_state_alphabet,
                purge_other_state_alphabets=purge_other_state_alphabets)

    def taxon_state_sets_map(self,
            char_indices=None,
            gaps_as_missing=True,
            gap_state=None,
            no_data_state=None):
        """
        Returns a dictionary that maps taxon objects to lists of sets of
        fundamental state indices.

        Parameters
        ----------

        char_indices : iterable of ints
            An iterable of indexes of characters to include (by column). If not
            given or |None| [default], then all characters are included.

        gaps_as_missing : boolean
            If |True| [default] then gap characters will be treated as missing
            data values. If |False|, then they will be treated as an additional
            (fundamental) state.`

        Returns
        -------
        d : dict
            A dictionary with class:|Taxon| objects as keys and a list of sets
            of fundamental state indexes as values.

            E.g., Given the following matrix of DNA characters:

                T1 AGN
                T2 C-T
                T3 GC?

            Return with ``gaps_as_missing==True`` ::

                {
                    <T1> : [ set([0]), set([2]),        set([0,1,2,3]) ],
                    <T2> : [ set([1]), set([0,1,2,3]),  set([3]) ],
                    <T3> : [ set([2]), set([1]),        set([0,1,2,3]) ],
                }

            Return with ``gaps_as_missing==False`` ::

                {
                    <T1> : [ set([0]), set([2]),        set([0,1,2,3]) ],
                    <T2> : [ set([1]), set([4]),        set([3]) ],
                    <T3> : [ set([2]), set([1]),        set([0,1,2,3,4]) ],
                }

            Note that when gaps are treated as a fundamental state, not only
            does '-' map to a distinct and unique state (4), but '?' (missing
            data) maps to set consisting of all bases *and* the gap
            state, whereas 'N' maps to a set of all bases but not including the
            gap state.

            When gaps are treated as missing, on the other hand, then '?' and
            'N' and '-' all map to the same set, i.e. of all the bases.

        """
        taxon_to_state_indices = {}
        for t in self:
            cdv = self[t]
            if char_indices is None:
                ci = range(len(cdv))
            else:
                ci = char_indices
            v = []
            for char_index in ci:
                state = cdv[char_index]
                if gaps_as_missing:
                    v.append(set(state.fundamental_indexes_with_gaps_as_missing))
                else:
                    v.append(set(state.fundamental_indexes))
            taxon_to_state_indices[t] = v
        return taxon_to_state_indices

    def folded_site_frequency_spectrum(self, is_pad_vector_to_unfolded_length=False):
        r"""
        Returns the folded or minor site/allele frequency spectrum.

        Given $N$ chromosomes, the site frequency spectrum is a vector $(f_0,
        f_1, f_2, ..., f_N)$, where the value $f_i$ is the number of
        sites where $i$ derived alleles are segregating in the sample: 0
        alleles, 1 allele, 2 alleles, etc.

        The *folded* site frequency spectrum is a vector $(f_0, f_1, f_2, ...,
        f_m), m = \ceil{\frac{N}{2}}$, where the values are the number of minor
        alleles in the site.

        Parameters
        ----------
        is_pad_vector_to_unfolded_length: bool
            If False, then the vector length will be $\ceil{\frac{N}{2}}$,
            where $N$ is the number of taxa. Otherwise, by default,
            True, length of vector will be number of taxa + 1, with the
            first element the number of monomorphic sites not contributing to
            the site frequency spectrum.

        Returns
        -------
        v : list[int]
            A vector of integers representing the folded site frequency
            spectrum.
        """
        site_columns = zip(*self.sequences())
        nsites = 0
        if is_pad_vector_to_unfolded_length:
            sfs = [0 for idx in range(len(self._taxon_sequence_map)+1)]
        else:
            sfs = [0 for idx in range(int(math.ceil(len(self._taxon_sequence_map)/2.0))+1)]
        for site in site_columns:
            counter = collections.Counter(site)
            nsites += 1
            if len(counter) == 1:
                sfs[0] += 1
                continue
            del counter[counter.most_common(1)[0][0]]
            sfs[sum(counter.values())] += 1
        assert sum(sfs) == nsites
        return sfs

### Fixed Alphabet Characters ##################################################

class FixedAlphabetCharacterDataSequence(CharacterDataSequence):
    pass

class FixedAlphabetCharacterMatrix(DiscreteCharacterMatrix):

    character_sequence_type = FixedAlphabetCharacterDataSequence
    data_type = "fixed"
    datatype_alphabet = None

    def __init__(self, *args, **kwargs):
        DiscreteCharacterMatrix.__init__(self, *args, **kwargs)
        self.state_alphabets.append(self.__class__.datatype_alphabet)
        self._default_state_alphabet = self.__class__.datatype_alphabet

    def coerce_values(self, values):
        if self.datatype_alphabet is None:
            raise ValueError("'datatype_alphabet' not set")
        return charstatemodel.coerce_to_state_identities(
                state_alphabet=self.datatype_alphabet,
                values=values)

### DNA Characters ##################################################

class DnaCharacterDataSequence(FixedAlphabetCharacterDataSequence):
    pass

class DnaCharacterMatrix(FixedAlphabetCharacterMatrix):
    """
    Specializes |CharacterMatrix| for DNA data.
    """
    character_sequence_type = DnaCharacterDataSequence
    data_type = "dna"
    datatype_alphabet = DNA_STATE_ALPHABET

### RNA Characters ##################################################

class RnaCharacterDataSequence(FixedAlphabetCharacterDataSequence):
    pass

class RnaCharacterMatrix(FixedAlphabetCharacterMatrix):
    """
    Specializes |CharacterMatrix| for DNA data.
    """
    character_sequence_type = RnaCharacterDataSequence
    data_type = "rna"
    datatype_alphabet = RNA_STATE_ALPHABET

### Nucleotide Characters ##################################################

class NucleotideCharacterDataSequence(FixedAlphabetCharacterDataSequence):
    pass

class NucleotideCharacterMatrix(FixedAlphabetCharacterMatrix):
    """
    Specializes |CharacterMatrix| for RNA data.
    """
    character_sequence_type = NucleotideCharacterDataSequence
    data_type = "nucleotide"
    datatype_alphabet = NUCLEOTIDE_STATE_ALPHABET

### Protein Characters ##################################################

class ProteinCharacterDataSequence(FixedAlphabetCharacterDataSequence):
    pass

class ProteinCharacterMatrix(FixedAlphabetCharacterMatrix):
    """
    Specializes |CharacterMatrix| for protein or amino acid data.
    """
    character_sequence_type = ProteinCharacterDataSequence
    data_type = "protein"
    datatype_alphabet = PROTEIN_STATE_ALPHABET

### Restricted Site Characters ##################################################

class RestrictionSitesCharacterDataSequence(FixedAlphabetCharacterDataSequence):
    pass

class RestrictionSitesCharacterMatrix(FixedAlphabetCharacterMatrix):
    """
    Specializes |CharacterMatrix| for restriction site data.
    """
    character_sequence_type = RestrictionSitesCharacterDataSequence
    data_type = "restriction"
    datatype_alphabet = RESTRICTION_SITES_STATE_ALPHABET

### Infinite Sites Characters ##################################################

class InfiniteSitesCharacterDataSequence(FixedAlphabetCharacterDataSequence):
    pass

class InfiniteSitesCharacterMatrix(FixedAlphabetCharacterMatrix):
    """
    Specializes |CharacterMatrix| for infinite sites data.
    """
    character_sequence_type = InfiniteSitesCharacterDataSequence
    data_type = "infinite"
    datatype_alphabet = INFINITE_SITES_STATE_ALPHABET

### Standard Characters ##################################################

class StandardCharacterDataSequence(DiscreteCharacterDataSequence):
    pass

class StandardCharacterMatrix(DiscreteCharacterMatrix):
    """
    Specializes |CharacterMatrix| for "standard" data (i.e., generic discrete
    character data).

    """
    character_sequence_type = StandardCharacterDataSequence

    data_type = "standard"

    def __init__(self, *args, **kwargs):
        """
        A default state alphabet consisting of state symbols of 0-9 will
        automatically be created unless the ``default_state_alphabet=None`` is
        passed in. To specify a different default state alphabet::

            default_state_alphabet=dendropy.new_standard_state_alphabet("abc")
            default_state_alphabet=dendropy.new_standard_state_alphabet("ij")

        """
        if "default_state_alphabet" in kwargs:
            default_state_alphabet = kwargs.pop("default_state_alphabet")
        else:
            default_state_alphabet = charstatemodel.new_standard_state_alphabet()
        DiscreteCharacterMatrix.__init__(self, *args, **kwargs)
        if default_state_alphabet is not None:
            self.default_state_alphabet = default_state_alphabet

    def coerce_values(self, values):
        if self.default_state_alphabet is None:
            raise ValueError("'default_state_alphabet' not set")
        return charstatemodel.coerce_to_state_identities(
                state_alphabet=self.default_state_alphabet,
                values=values)


###############################################################################
## Main Character Matrix Factory Function

data_type_matrix_map = {
    'continuous' : ContinuousCharacterMatrix,
    'dna' : DnaCharacterMatrix,
    'rna' : RnaCharacterMatrix,
    'nucleotide' : NucleotideCharacterMatrix,
    'protein' : ProteinCharacterMatrix,
    'standard' : StandardCharacterMatrix,
    'restriction' : RestrictionSitesCharacterMatrix,
    'infinite' : InfiniteSitesCharacterMatrix,
}

def get_char_matrix_type(data_type):
    if data_type is None:
        raise TypeError("'data_type' must be specified")
    matrix_type = data_type_matrix_map.get(data_type, None)
    if matrix_type is None:
        raise KeyError("Unrecognized data type specification: '{}'".format(data_type,
            sorted(data_type_matrix_map.keys())))
    return matrix_type

def new_char_matrix(data_type, **kwargs):
    matrix_type = get_char_matrix_type(data_type=data_type)
    m = matrix_type(**kwargs)
    return m



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Writing of Newick-format tree to a stream.
"""

import re
import warnings
from dendropy.utility import error
from dendropy.utility import textprocessing
from dendropy.dataio import tokenizer
from dendropy.dataio import nexusprocessing
from dendropy.dataio import ioservice

##############################################################################
## NewickReader

class NewickWriter(ioservice.DataWriter):
    """
    Formatter for Newick data.
    """

    def __init__(self, **kwargs):
        """

        Keyword Arguments
        -----------------
        suppress_leaf_taxon_labels : boolean, default: |False|
            If |True|, then taxon labels will not be rendered for leaves.
            Default is |False|: render leaf taxon labels. See notes below for
            details.
        suppress_leaf_node_labels : boolean, default: |True|
            If |False|, then node labels (if available) will be printed for
            leaves. Defaults to |True|: do not render leaf node labels. See
            notes below for details.
        suppress_internal_taxon_labels : boolean, default: |False|
            If |True|, then taxon labels will not be printed for internal
            nodes. Default is |False|: print taxon labels for internal nodes.
            See notes below for details.
        suppress_internal_node_labels : boolean, default: |False|
            If |True|, then node labels will not be printed for internal nodes.
            Default is |False|: print node labels for internal nodes. See notes
            below for details.
        suppress_rooting : boolean, default: |False|
            If |True|, will not write rooting token ('[&R]' or '[&U]').
            Default is |False|: rooting token will be written.
        suppress_edge_lengths : boolean, default: |False|
            If |True|, will not write edge lengths. Default is |False|: edge
            lengths will be written.
        unquoted_underscores : boolean, default: |False|
            If |True|, labels with underscores will not be quoted, which will
            mean that they will be interpreted as spaces if read again ("soft"
            underscores).  If |False|, then labels with underscores
            will be quoted, resulting in "hard" underscores.  Default is
            |False|.
        preserve_spaces : boolean, default: |False|
            If |True|, spaces will not be replaced with underscores in labels
            (which means any labels containing spaces will have to be quoted).
            Default is |False|: spaces will be converted to underscores.
            False.
        store_tree_weights : boolean, default: |False|
            If |True|, tree weights are written. Default is |False|: tree
            weights will not be written.
        taxon_token_map : boolean or dict or |None|, default: |None|.
            If not |False| or |None|, a "TRANSLATE" statement will be written
            and referenced in tree statements (instead of using the taxon
            labels). If |True|, then a default translate statement will
            be used, with tokens given by the taxon indexes. If a dictionary is
            given, then the keys should be |Taxon| objects and the
            values should be the token (strings).
        suppress_annotations : boolean, default: |True|
            If |False|, metadata annotations will be written out as special
            comments. Defaults to |True|: metadata annotations will be ignored.
        annotations_as_nhx : boolean, default: |False|
            If |True|, and if ``suppress_annotations`` is |False|, will write
            annotation as NHX statements. Default is |False|: annotations
            will not be written as NHX statements.
        suppress_item_comments : boolean, default: |True|
            If |False|, any additional comments associated with trees, nodes,
            edges, etc. will be written. Default is |True|: comments will be
            ignored.
        node_label_element_separator : string, default: ' '
            If both ``suppress_leaf_taxon_labels`` and
            ``suppress_leaf_node_labels`` are |False|, then this will be the
            string used to join them. Defaults to ' ' (space).
        node_label_compose_fn : function object or |None|, default: |None|
            If not |None|, should be a function that takes a |Node|
            object as an argument and returns the string to be used to
            represent the node in the tree statement. The return value from
            this function is used unconditionally to print a node
            representation in a tree statement, by-passing the default
            labelling function, ignoring ``suppress_leaf_taxon_labels``,
            ``suppress_leaf_node_labels=True``, ``suppress_internal_taxon_labels``,
            ``suppress_internal_node_labels``, etc. Defaults to |None|.
        edge_label_compose_fn : function object or |None|, default: |None|
            If not |None|, should be a function that takes an Edge object as
            an argument, and returns the string to be used to represent the
            edge length in the tree statement.
        real_value_format_specifier : string, default: ''
            Format specification for real/float values. Will be applied to edge
            lengths (if ``edge_label_compose_fn`` is not given) as well as
            annotations. The format specifier should be given in Python's
            string format specification mini-language. E.g. ".8f", ".4E",
            "8.4f".
        ignore_unrecognized_keyword_arguments : boolean, default: |False|
            If |True|, then unsupported or unrecognized keyword arguments will
            not result in an error. Default is |False|: unsupported keyword
            arguments will result in an error.

        Notes
        -----

        DendroPy distinguishes between *taxon* labels and *node*
        labels.

        In a Newick string, however, no such distinction is possible, and
        any one node can only be rendered with a single token or symbol. Thus,
        if there is more than one source of label available for a particular
        node (e.g., if both ``suppress_leaf_taxon_labels`` and
        ``suppress_leaf_node_labels`` are |False|, and a particular leaf
        node has both a taxon *and* a label associated with it), then
        the node symbol will be rendered as concatenation of the unsuppressed
        candidate labels, with each candidate label separated by the value
        given in ``node_label_element_separator``. Note that this concatenated
        label requires special handling when being re-read to avoid being
        interpreted as the operational taxonomic unit concept label in
        its entirety. These defaults can all be overridden using the
        various keywords, or a custom label can be composed for the
        node by passing an appropriate function object via the
        ``node_label_compose_fn`` argument.

        Note that, in typical Newick usage, labels of leaf nodes represent
        operational taxonomic unit concepts, and thus the default setting to
        render leaf taxon labels but suppress leaf node labels. Internal node
        labels, on the other hand, are typically used both to represent
        operational taxonomic unit concepts (e.g., ancestral taxa) as well as
        other concepts (e.g., support or geographic range), and thus the
        default internal node rendering is to not to suppress either the taxon
        labels or the node labels.

        """
        ioservice.DataWriter.__init__(self)
        legacy = {
                "internal_labels": "Use 'suppress_internal_taxon_labels' instead",
                "write_rooting": "Use 'suppress_rooting' instead",
                "quote_undescores" : "Use 'unquoted_underscores' instead",
                "annotations_as_comments": "Use 'suppress_annotations' instead"
                }
        for kw in legacy:
            if kw in kwargs:
                raise TypeError("'{}' is no longer supported: {}".format(kw, legacy[kw]))
        ioservice.DataWriter.__init__(self, **kwargs)
        self.suppress_leaf_taxon_labels = kwargs.pop("suppress_leaf_taxon_labels", False)
        self.suppress_leaf_node_labels = kwargs.pop("suppress_leaf_node_labels", True)
        self.suppress_internal_taxon_labels = kwargs.pop("suppress_internal_taxon_labels", False)
        # self.suppress_internal_taxon_labels = not kwargs.pop("internal_labels", not sef.suppress_internal_taxon_labels) # legacy
        self.suppress_internal_node_labels = kwargs.pop("suppress_internal_node_labels", False)
        # self.suppress_internal_node_labels = not kwargs.pop("internal_labels", not self.suppress_internal_node_labels) # legacy
        self.suppress_rooting = kwargs.pop("suppress_rooting", False)
        # self.suppress_rooting = not kwargs.pop("write_rooting", not self.suppress_rooting) # legacy
        self.suppress_edge_lengths = kwargs.pop("suppress_edge_lengths", False)
        # self.suppress_edge_lengths = not kwargs.pop("edge_lengths", not self.suppress_edge_lengths) # legacy
        self.unquoted_underscores = kwargs.pop("unquoted_underscores", False)
        # self.unquoted_underscores = not kwargs.pop('quote_underscores', not self.unquoted_underscores) # legacy
        self.preserve_spaces = kwargs.pop("preserve_spaces", False)
        self.store_tree_weights = kwargs.pop("store_tree_weights", False)
        self.taxon_token_map = kwargs.pop("taxon_token_map", {})
        self.suppress_annotations = kwargs.pop("suppress_annotations", True)
        # self.suppress_annotations = not kwargs.pop("annotations_as_comments", not self.suppress_annotations) # legacy
        self.annotations_as_nhx = kwargs.pop("annotations_as_nhx", False)
        self.suppress_item_comments = kwargs.pop("suppress_item_comments", True)
        self.suppress_item_comments = not kwargs.pop("write_item_comments", not self.suppress_item_comments)
        self.node_label_element_separator = kwargs.pop("node_label_element_separator", ' ')
        self.node_label_compose_fn = kwargs.pop("node_label_compose_fn", None)
        self.edge_label_compose_fn = kwargs.pop("edge_label_compose_fn", None)
        self._real_value_format_specifier = ""
        self._real_value_formatter = None
        self.real_value_format_specifier = kwargs.pop("real_value_format_specifier", self._real_value_format_specifier)
        if self.edge_label_compose_fn is None:
            self.edge_label_compose_fn = self._format_edge_length
        self.check_for_unused_keyword_arguments(kwargs)

    def _get_taxon_tree_token(self, taxon):
        if self.taxon_token_map is None:
            self.taxon_token_map = {}
        try:
            return self.taxon_token_map[taxon]
        except KeyError:
            t = str(taxon.label)
            self.taxon_token_map[taxon] = t
            return t

    def _get_real_value_format_specifier(self):
        return self._real_value_format_specifier
    def _set_real_value_format_specifier(self, f):
        if f is None:
            f = ""
        self._real_value_format_specifier = f
        s = "{:" + self._real_value_format_specifier + "}"
        self._real_value_formatter = s.format
    real_value_format_specifier = property(_get_real_value_format_specifier, _set_real_value_format_specifier)

    def _format_edge_length(self, edge):
        """
        Note: instance method to allow overriding.
        """
        return self._real_value_formatter(edge.length)

    def _write(self,
            stream,
            taxon_namespaces=None,
            tree_lists=None,
            char_matrices=None,
            global_annotations_target=None):
        for tree_list in tree_lists:
            if (self.attached_taxon_namespace is not None
                    and tree_list.taxon_namespace is not self.attached_taxon_namespace):
                continue
            self._write_tree_list(stream, tree_list)

    def _write_tree_list(self, stream, tree_list):
        """
        Writes a |TreeList| in Newick schema to ``stream``.
        """
        for tree in tree_list:
            self._write_tree(stream, tree)
            stream.write("\n")
        # In Newick format, no clear way to distinguish between
        # annotations/comments associated with tree collection and
        # annotations/comments associated with first tree. So we place them at
        # *end* of document.
        if (not self.suppress_annotations) and (hasattr(tree_list, "_annotations")):
            annotation_comments = nexusprocessing.format_item_annotations_as_comments(tree_list,
                    nhx=self.annotations_as_nhx,
                    real_value_format_specifier=self.real_value_format_specifier,
                    )
        else:
            annotation_comments = ""
        treelist_comments = self._compose_comment_string(tree_list)
        stream.write("{}{}".format(
                annotation_comments,
                treelist_comments))

    def _write_tree(self, stream, tree):
        """
        Composes and writes ``tree`` to ``stream``.
        """
        if tree.rooting_state_is_undefined or self.suppress_rooting:
            rooting = ""
        elif tree.is_rooted:
            rooting = "[&R] "
        elif not tree.is_rooted:
            rooting = "[&U] "
        else:
            rooting = ""
        if self.store_tree_weights and tree.weight is not None:
            weight = "[&W {}] ".format(tree.weight)
        else:
            weight = ""
        if not self.suppress_annotations:
            annotation_comments = nexusprocessing.format_item_annotations_as_comments(tree,
                    nhx=self.annotations_as_nhx,
                    real_value_format_specifier=self.real_value_format_specifier,
                    )
        else:
            annotation_comments = ""
        tree_comments = self._compose_comment_string(tree)
        stream.write("{}{}{}{}".format(
                rooting,
                weight,
                annotation_comments,
                tree_comments,
                ))
        tree.apply(
                before_fn=lambda x: self._write_node_open(x, stream),
                after_fn=lambda x: self._write_node_close(x, stream),
                leaf_fn=lambda x: self._write_leaf(x, stream),
                )
        stream.write(";")

    def _write_node_open(self, node, out):
        if node._parent_node is None or node._parent_node._child_nodes[0] is node:
            out.write("(")
        else:
            out.write(",(")

    def _write_leaf(self, node, out):
        if not (node._parent_node is None or node._parent_node._child_nodes[0] is node):
            out.write(",")
        self._write_node_body(node, out)

    def _write_node_close(self, node, out):
        out.write(")")
        self._write_node_body(node, out)

    def _write_node_body(self, node, out):
        out.write(self._render_node_tag(node))
        if node.edge and node.edge.length != None and not self.suppress_edge_lengths:
            out.write(":{}".format(self.edge_label_compose_fn(node.edge)))
        if not self.suppress_annotations:
            node_annotation_comments = nexusprocessing.format_item_annotations_as_comments(node,
                    nhx=self.annotations_as_nhx,
                    real_value_format_specifier=self.real_value_format_specifier)
            out.write(node_annotation_comments)
            edge_annotation_comments = nexusprocessing.format_item_annotations_as_comments(node.edge,
                    nhx=self.annotations_as_nhx,
                    real_value_format_specifier=self.real_value_format_specifier)
            out.write(edge_annotation_comments)
        out.write(self._compose_comment_string(node))
        out.write(self._compose_comment_string(node.edge))

    def _compose_comment_string(self, item):
        if not self.suppress_item_comments and item.comments:
            item_comments = []
            if textprocessing.is_str_type(item.comments):
                item.comments = [item.comments]
            for comment in item.comments:
                item_comments.append("[{}]".format(comment))
            item_comment_str = "".join(item_comments)
        else:
            item_comment_str = ""
        return item_comment_str

    def _render_node_tag(self, node):
        """
        Based on current settings, the attributes of a node, and
        whether or not the node is a leaf, returns an appropriate tag.
        """
        tag = None
        if self.node_label_compose_fn:
            tag = self.node_label_compose_fn(node)
        else:
            tag_parts = []
            is_leaf = len(node.child_nodes()) == 0
            if is_leaf:
                if hasattr(node, 'taxon') \
                        and node.taxon \
                        and node.taxon.label is not None \
                        and not self.suppress_leaf_taxon_labels:
                    tag_parts.append(self._get_taxon_tree_token(node.taxon))
                if hasattr(node, 'label') \
                        and node.label \
                        and node.label is not None \
                        and not self.suppress_leaf_node_labels:
                    tag_parts.append(str(node.label))
                if len(tag_parts) > 0:
                    tag = self.node_label_element_separator.join(tag_parts)
                else:
                    return "" # anonymous leaf
            else:
                if hasattr(node, 'taxon') \
                        and node.taxon \
                        and node.taxon.label is not None \
                        and not self.suppress_internal_taxon_labels:
                    tag_parts.append(self._get_taxon_tree_token(node.taxon))
                if hasattr(node, 'label') \
                        and node.label \
                        and node.label is not None \
                        and not self.suppress_internal_node_labels:
                    tag_parts.append(str(node.label))
                if len(tag_parts) > 0:
                    tag = self.node_label_element_separator.join(tag_parts)
                else:
                    return ""
        if tag:
            tag = nexusprocessing.escape_nexus_token(tag,
                    preserve_spaces=self.preserve_spaces,
                    quote_underscores=not self.unquoted_underscores)
            return tag
        else:
            return ""

    # def _compose_node(self, node):
    #     """
    #     Given a DendroPy Node, this returns the Node as a Newick
    #     statement according to the class-defined formatting rules.
    #     """
    #     child_nodes = node.child_nodes()
    #     if child_nodes:
    #         subnodes = [self._compose_node(child) for child in child_nodes]
    #         statement = '(' + ','.join(subnodes) + ')'
    #         if not (self.suppress_internal_taxon_labels and self.suppress_internal_node_labels):
    #             statement = statement + self._render_node_tag(node)
    #         if node.edge and node.edge.length != None and not self.suppress_edge_lengths:
    #             statement =  "{}:{}".format(statement, self.edge_label_compose_fn(node.edge))
    #     else:
    #         statement = self._render_node_tag(node)
    #         if node.edge and node.edge.length != None and not self.suppress_edge_lengths:
    #             statement =  "{}:{}".format(statement, self.edge_label_compose_fn(node.edge))
    #     if not self.suppress_annotations:
    #         node_annotation_comments = nexusprocessing.format_item_annotations_as_comments(node,
    #                 nhx=self.annotations_as_nhx,
    #                 real_value_format_specifier=self.real_value_format_specifier)
    #         edge_annotation_comments = nexusprocessing.format_item_annotations_as_comments(node.edge,
    #                 nhx=self.annotations_as_nhx,
    #                 real_value_format_specifier=self.real_value_format_specifier)
    #         statement = statement + node_annotation_comments + edge_annotation_comments
    #     edge_comment_str = self._compose_comment_string(node.edge)
    #     node_comment_str = self._compose_comment_string(node)
    #     statement = statement + node_comment_str + edge_comment_str
    #     return statement



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
DEPRECATED IN DENDROPY 4: USE `dendropy.simulate.treesim` instead.
"""

from dendropy.simulate import treesim
from dendropy.simulate import popgensim
from dendropy.utility import deprecate

def pop_gen_tree(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.pop_gen_tree()' function has moved to 'dendropy.simulate.popgensim.pop_gen_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.pop_gen_tree(...)",
            new_construct="from dendropy.simulate import popgensim\ntree = popgensim.pop_gen_tree(...)")
    return popgensim.pop_gen_tree(*args, **kwargs)

def birth_death(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.birth_death()' function has moved to 'dendropy.simulate.treesim.birth_death_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.birth_death(...)",
            new_construct="from dendropy.simulate import treesim\ntree = treesim.birth_death_tree(...)")
    return treesim.birth_death_tree(*args, **kwargs)

# from dendropy.model.birthdeath import discrete_birth_death_tree as discrete_birth_death
def discrete_birth_death(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.discrete_birth_death()' function has moved to 'dendropy.simulate.treesim.discrete_birth_death_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.discrete_birth_death(...)",
            new_construct="from dendropy.simulate import treesim\ntree = treesim.discrete_birth_death_tree(...)")
    return treesim.discrete_birth_death_tree(*args, **kwargs)

# from dendropy.model.birthdeath import uniform_pure_birth_tree as uniform_pure_birth
def uniform_pure_birth(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.uniform_pure_birth()' function has moved to 'dendropy.simulate.treesim.uniform_pure_birth_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.uniform_pure_birth(...)",
            new_construct="from dendropy.simulate import treesim\ntree = treesim.uniform_pure_birth_tree(...)")
    return treesim.uniform_pure_birth_tree(*args, **kwargs)

# from dendropy.model.coalescent import contained_coalescent_tree as contained_coalescent
def contained_coalescent(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.contained_coalescent()' function has moved to 'dendropy.simulate.treesim.contained_coalescent_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.contained_coalescent(...)",
            new_construct="from dendropy.simulate import treesim\ntree = treesim.contained_coalescent_tree(...)")
    return treesim.contained_coalescent_tree(*args, **kwargs)

# from dendropy.model.coalescent import pure_kingman_tree as pure_kingman
def pure_kingman(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.pure_kingman()' function has moved to 'dendropy.simulate.treesim.pure_kingman_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.pure_kingman(...)",
            new_construct="from dendropy.simulate import treesim\ntree = treesim.pure_kingman_tree(...)")
    return treesim.pure_kingman_tree(*args, **kwargs)

# from dendropy.model.coalescent import mean_kingman_tree as mean_kingman
def mean_kingman(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.mean_kingman()' function has moved to 'dendropy.simulate.treesim.mean_kingman_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.mean_kingman(...)",
            new_construct="from dendropy.simulate import treesim\ntree = treesim.mean_kingman_tree(...)")
    return treesim.mean_kingman_tree(*args, **kwargs)

# from dendropy.model.coalescent import constrained_kingman_tree as constrained_kingman
def constrained_kingman(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.constrained_kingman()' function has moved to 'dendropy.simulate.treesim.constrained_kingman_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.constrained_kingman(...)",
            new_construct="from dendropy.simulate import treesim\ntree = treesim.constrained_kingman_tree(...)")
    return treesim.constrained_kingman_tree(*args, **kwargs)

# from dendropy.model.treeshape import star_tree
def star_tree(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.treesim.star()' function has moved to 'dendropy.simulate.treesim.star_tree()'.",
            old_construct="from dendropy import treesim\ntree = treesim.star(...)",
            new_construct="from dendropy.simulate import treesim\ntree = treesim.star_tree(...)")
    return treesim.star_tree(*args, **kwargs)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.txt" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Profile distances.
"""

import math
import collections
from dendropy.utility import constants
from dendropy.model import coalescent

class MeasurementProfile(object):

    @staticmethod
    def _euclidean_distance(v1, v2, is_weight_values_by_comparison_profile_size=True):
        v1_size = len(v1)
        v2_size = len(v2)
        v1_idx = 0
        v2_idx = 0
        if v1_size > v2_size:
            v1_idx = v1_size - v2_size
            weight = float(v2_size)
        elif v2_size > v1_size:
            v2_idx = v2_size - v1_size
            weight = float(v1_size)
        else:
            weight = float(v1_size)
        if not is_weight_values_by_comparison_profile_size:
            weight = 1.0
        ss = 0.0
        while v1_idx < v1_size and v2_idx < v2_size:
            ss += pow(v1[v1_idx]/weight - v2[v2_idx]/weight, 2)
            v1_idx += 1
            v2_idx += 1
        return math.sqrt(ss)

    def __init__(self,
            profile_data=None,
            interpolation_method="piecewise_linear"):
        self.set_data(profile_data)
        self.fixed_size = 0
        self.interpolation_method = interpolation_method

    def add(self, value):
        self._profile_data.append(value)
        self._profile_data = sorted(self._profile_data)
        self._raw_data_size = len(self._profile_data)
        self._interpolated_profiles = {}

    def set_data(self, values):
        if values is None:
            values = []
        self._profile_data = sorted(values)
        self._raw_data_size = len(self._profile_data)
        self._interpolated_profiles = {}

    def __len__(self):
        return self._raw_data_size

    def distance(self,
            other,
            profile_size,
            is_weight_values_by_comparison_profile_size=True):
        if profile_size is None:
            profile_size = self._get_profile_comparison_size(other)
        v1 = self._get_profile_for_size(profile_size)
        v2 = other._get_profile_for_size(profile_size)
        return MeasurementProfile._euclidean_distance(v1, v2,
                is_weight_values_by_comparison_profile_size=is_weight_values_by_comparison_profile_size)
    # def get(self, idx):
    #     if idx >= self._raw_data_size:
    #         val = 0.0
    #     else:
    #         val = self._profile_data[idx]
    #         if val is None:
    #             val = 0.0
    #     return val

    def _get_profile_comparison_size(self, other):
        if self.fixed_size > 0 and other.fixed_size > 0:
            if self.fixed_size != other.fixed_size:
                raise ValueError("Comparing two profiles locked to different sizes: {} and {}".format(
                    self.fixed_size, other.fixed_size))
            return self.fixed_size
        elif self.fixed_size == 0 and other.fixed_size > 0:
            if other.fixed_size < self._raw_data_size:
                raise ValueError("Cannot interpolate points in current profile: current raw data size is {} but other profile is locked to a smaller fixed size ({})".format(
                    self._raw_data_size, other.fixed_size))
            return other.fixed_size
        elif self.fixed_size > 0 and other.fixed_size == 0:
            if self.fixed_size < other._raw_data_size:
                raise ValueError("Cannot interpolate points in other profile: other raw data size is {} but current profile is locked to a smaller fixed size ({})".format(
                    other._raw_data_size,
                    self.fixed_size,
                    ))
            return self.fixed_size
        else:
            return max(self._raw_data_size, other._raw_data_size)

    def _get_profile_for_size(self, profile_size):
        if not profile_size:
            return self._raw_data_size
        try:
            return self._interpolated_profiles[profile_size]
        except KeyError:
            self._interpolated_profiles[profile_size] = self._generate_interpolated_profile(profile_size)
            return self._interpolated_profiles[profile_size]

    def _generate_interpolated_profile(self, profile_size):
        if profile_size == self._raw_data_size:
            self._interpolated_profiles[profile_size] = list(self._profile_data)
            return self._interpolated_profiles[profile_size]
        if self._raw_data_size == 0:
            raise ValueError("No data in profile")
        if profile_size < self._raw_data_size:
            raise ValueError("Error interpolating points in profile: number of requested interpolated points ({}) is less than raw data size ({})".format(
                profile_size, self._raw_data_size))
        default_bin_size = int(profile_size / self._raw_data_size)
        if default_bin_size == 0:
            raise ValueError("Profile size ({}) is too small for raw data size ({}), resulting in a null bin size".
                    format(profile_size, self._raw_data_size))
        bin_sizes = [default_bin_size] * self._raw_data_size

        # // due to rounding error, default bin size may not be enough
        # // this hacky algorithm adds additional points to bins to make up the
        # // difference, trying to distribute the additional points along the
        # // length of the line
        diff = profile_size - (default_bin_size * self._raw_data_size)
        if diff > 0:
            dv = float(diff) / self._raw_data_size
            cv = 0.0
            for bin_idx in range(len(bin_sizes)):
                if diff <= 1.0:
                    break
                cv += dv
                if cv >= 1.0:
                    bin_sizes[bin_idx] += 1
                    diff -= 1.0
                    cv = cv - 1.0

        interpolated_profile = []
        if self.interpolation_method == "staircase":
            for original_data_value in self._profile_data:
                self._interpolate_flat(
                        interpolated_profile=interpolated_profile,
                        value=original_data_value,
                        num_points=default_bin_size)
        elif self.interpolation_method == "piecewise_linear":
            for bin_idx, original_data_value in enumerate(self._profile_data[:-1]):
                self._interpolate_linear(
                        interpolated_profile=interpolated_profile,
                        x1=bin_idx,
                        y1=self._profile_data[bin_idx],
                        y2=self._profile_data[bin_idx+1],
                        num_points=bin_sizes[bin_idx],
                        max_points=profile_size,
                        )
            interpolated_profile.append(self._profile_data[-1])
        return interpolated_profile

    def _interpolate_flat(self,
            interpolated_profile,
            value,
            num_points):
        interpolated_profile += [value] * num_points
        return interpolated_profile

    def _interpolate_linear(self,
            interpolated_profile,
            x1, y1, y2,
            num_points,
            max_points):
        assert num_points > 0
        slope = float(y2 - y1) / num_points
        for xi in range(num_points):
            if max_points and len(interpolated_profile) >= max_points:
                return interpolated_profile
            interpolated_profile.append((slope * xi) + y1)

class TreeProfile(object):

    def __init__(self,
            tree,
            is_measure_edge_lengths=True,
            is_measure_patristic_distances=False,
            is_measure_patristic_steps=False,
            is_measure_node_distances=True,
            is_measure_node_steps=True,
            is_measure_node_ages=True,
            is_measure_coalescence_intervals=True,
            is_normalize=True,
            ultrametricity_precision=constants.DEFAULT_ULTRAMETRICITY_PRECISION,
            tree_phylogenetic_distance_matrix=None,
            tree_node_distance_matrix=None,
            tree_id=None,
            is_skip_normalization_on_zero_division_error=False,
            ):
        self.tree_id = tree_id
        self.is_measure_edge_lengths = is_measure_edge_lengths
        self.is_measure_patristic_distances = is_measure_patristic_distances
        self.is_measure_patristic_steps = is_measure_patristic_steps
        self.is_measure_node_distances = is_measure_node_distances
        self.is_measure_node_steps = is_measure_node_steps
        self.is_measure_node_ages = is_measure_node_ages
        self.is_measure_coalescence_intervals = is_measure_coalescence_intervals
        self.is_normalize = is_normalize
        self.is_skip_normalization_on_zero_division_error = is_skip_normalization_on_zero_division_error
        self.ultrametricity_precision = ultrametricity_precision
        self.measurement_profiles = collections.OrderedDict()
        self.compile(tree)

    def compile(self, tree,
            tree_phylogenetic_distance_matrix=None,
            tree_node_distance_matrix=None,
            ):
        if self.is_measure_edge_lengths:
            if self.is_normalize:
                tree_length_nf = tree.length()
                if tree_length_nf == 0:
                    tree_length_nf = 1.0
            else:
                tree_length_nf = 1.0
            self.measurement_profiles["Edge.Lengths"] = MeasurementProfile(
                    profile_data=[float(e.length)/tree_length_nf for e in tree.postorder_edge_iter() if e.length is not None])
        if self.is_measure_patristic_distances or self.is_measure_patristic_steps:
            if tree_phylogenetic_distance_matrix is None:
                tree_phylogenetic_distance_matrix = tree.phylogenetic_distance_matrix()
            if self.is_measure_patristic_distances:
                self.measurement_profiles["Patristic.Distances"] = MeasurementProfile(
                        profile_data=tree_phylogenetic_distance_matrix.distances(
                            is_weighted_edge_distances=True,
                            is_normalize_by_tree_size=self.is_normalize,))
            if self.is_measure_patristic_steps:
                self.measurement_profiles["Patristic.Steps"] = MeasurementProfile(
                        profile_data=tree_phylogenetic_distance_matrix.distances(
                        is_weighted_edge_distances=False,
                        is_normalize_by_tree_size=self.is_normalize,))
        if self.is_measure_node_distances or self.is_measure_node_steps:
            if tree_node_distance_matrix is None:
                tree_node_distance_matrix = tree.node_distance_matrix()
            if self.is_measure_node_distances:
                self.measurement_profiles["Node.Distances"] = MeasurementProfile(
                        profile_data=tree_node_distance_matrix.distances(
                            is_weighted_edge_distances=True,
                            is_normalize_by_tree_size=self.is_normalize,))
            if self.is_measure_node_steps:
                self.measurement_profiles["Node.Steps"] = MeasurementProfile(
                        profile_data=tree_node_distance_matrix.distances(
                        is_weighted_edge_distances=False,
                        is_normalize_by_tree_size=self.is_normalize,))
        if self.is_measure_node_ages:
            node_ages = tree.calc_node_ages(ultrametricity_precision=self.ultrametricity_precision)
            if self.is_normalize:
                s = sum(node_ages)
                try:
                    normalized_node_ages = [a/s for a in node_ages]
                    node_ages = normalized_node_ages
                except ZeroDivisionError as e:
                    if self.is_skip_normalization_on_zero_division_error:
                        pass
                    else:
                        raise
            self.measurement_profiles["Node.Ages"] = MeasurementProfile(profile_data=node_ages,)
        if self.is_measure_coalescence_intervals:
            cf = coalescent.extract_coalescent_frames(
                    tree=tree,
                    ultrametricity_precision=self.ultrametricity_precision,)
            waiting_times = cf.values()
            if self.is_normalize:
                s = sum(waiting_times)
                try:
                    normalized_waiting_times = [w/s for w in waiting_times]
                    waiting_times = normalized_waiting_times
                except ZeroDivisionError as e:
                    if self.is_skip_normalization_on_zero_division_error:
                        pass
                    else:
                        raise
            self.measurement_profiles["Coalescence.Intervals"] = MeasurementProfile(profile_data=waiting_times,)

    @property
    def measurement_names(self):
        names = []
        if self.is_measure_edge_lengths:
            names.append("Edge.Lengths")
        if self.is_measure_patristic_distances:
            names.append("Patristic.Distances")
        if self.is_measure_patristic_steps:
            names.append("Patristic.Steps")
        if self.is_measure_node_distances:
            names.append("Node.Distances")
        if self.is_measure_node_steps:
            names.append("Node.Steps")
        if self.is_measure_node_ages:
            names.append("Node.Ages")
        if self.is_measure_coalescence_intervals:
            names.append("Coalescence.Intervals")
        return names

    def measure_distances(self, other_tree_profile,
            profile_size=None,
            is_weight_values_by_comparison_profile_size=True):
        d = collections.OrderedDict()
        for pm_name in self.measurement_profiles:
            p1 = self.measurement_profiles[pm_name]
            p2 = other_tree_profile.measurement_profiles[pm_name]
            d[pm_name] = p1.distance(p2,
                    profile_size=None,
                    is_weight_values_by_comparison_profile_size=is_weight_values_by_comparison_profile_size)
        return d




#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Wrappers for interacting with SeqGen. Originally part of PySeqGen.
"""

import subprocess
import uuid
import tempfile
import socket
import random
import os
import sys

from optparse import OptionGroup
from optparse import OptionParser

import dendropy
from dendropy.utility.messaging import get_logger
from dendropy.utility import processio
_LOG = get_logger("interop.seqgen")

HOSTNAME = socket.gethostname()
PID = os.getpid()

def _get_strongly_unique_tempfile(dir=None):
    return tempfile.NamedTemporaryFile(
            dir=dir,
            prefix="dendropy_tempfile-{0}-{1}-{2}".format(HOSTNAME, PID, uuid.uuid4()),
            mode="w",
            )

def _get_tempfile(dir=None):
    return tempfile.NamedTemporaryFile(
            dir=dir,
            mode="w")

class SeqGen(object):
    """
    This class wraps all attributes and input needed to make a call to SeqGen.
    """

    class SubstitutionModel(object):
        def __init__(self, idstr):
            self.idstr = idstr
        def __str__(self):
            return self.idstr

    F84 = SubstitutionModel("F84")
    HKY = SubstitutionModel("HKY")
    GTR = SubstitutionModel("GTR")
    JTT = SubstitutionModel("JTT")
    WAG = SubstitutionModel("WAG")
    PAM = SubstitutionModel("PAM")
    BLOSUM = SubstitutionModel("BLOSUM")
    MTREV = SubstitutionModel("MTREV")
    CPREV = SubstitutionModel("CPREV")
    GENERAL = SubstitutionModel("GENERAL")
    MODELS = [F84, HKY, GTR, JTT, WAG, PAM, BLOSUM, MTREV, CPREV, GENERAL]
    MODEL_IDS = [str(m) for m in MODELS]

    def get_model(idstr):
        for model in SeqGen.MODELS:
            if idstr.upper() == model.idstr.upper():
                return model
        return None
    get_model = staticmethod(get_model)

    def __init__(
        self,
        strongly_unique_tempfiles=False,
        rng=None,
        seq_len=None,
    ):
        """
        Sets up all properties, which (generally) map directly to command
        parameters of Seq-Gen.
        """

        # control tempfile generation
        if strongly_unique_tempfiles:
            self.get_tempfile = _get_strongly_unique_tempfile
        else:
            self.get_tempfile = _get_tempfile

        # python object specific attributes
        self.seqgen_path = 'seq-gen'
        self.rng_seed = None
        self._rng = rng

        # following are passed to seq-gen in one form or another
        self.char_model = 'HKY'
        self.seq_len = seq_len
        self.num_partitions = None
        self.scale_branch_lens = None
        self.scale_tree_len = None
        self.codon_pos_rates = None
        self.gamma_shape = None
        self.gamma_cats = None
        self.prop_invar = None
        self.state_freqs = None
        self.ti_tv = 0.5 # = kappa of 1.0, i.e. JC
        self.general_rates = None
        self.ancestral_seq_idx = None
        self.write_ancestral_seqs = False
        self.write_site_rates = False
        self.record_separator = "//"

    def _get_rng(self):
        if self._rng is None:
            self._rng = random.Random(self.rng_seed)
        return self._rng
    def _set_rng(self, rng):
        self._rng = rng
    rng = property(_get_rng, _set_rng)

    def _get_kappa(self):
        return float(self.ti_tv) / 2
    def _set_kappa(self, kappa):
        self.ti_tv = kappa * 2

    def _compose_arguments(
        self,
        output_format="nexus",
        append_file=None,
    ):
        """
        Composes and returns a list of strings that make up the arguments to a Seq-Gen
        call, based on the attribute values of the object.
        """
        args = []
        args.append(self.seqgen_path)
        if self.char_model:
            args.append("-m%s" % str(self.char_model))
        if self.seq_len:
            args.append("-l%s" % self.seq_len)
        if self.num_partitions:
            args.append("-p%s" % self.num_partitions)
        if self.scale_branch_lens:
            args.append("-s%s" % self.scale_branch_lens)
        if self.scale_tree_len:
            args.append("-d%s" % self.scale_tree_len)
        if self.codon_pos_rates:
            args.append("-c%s" % (",".join(self.codon_pos_rates)))
        if self.gamma_shape:
            args.append("-a%s" % self.gamma_shape)
        if self.gamma_cats:
            args.append("-g%s" % self.gamma_cats)
        if self.prop_invar:
            args.append("-i%s" % self.prop_invar)
        if self.state_freqs:
            if isinstance(self.state_freqs, str):
                args.append("-f%s" % self.state_freqs)
            else:
                args.append("-f%s" % (",".join([str(s) for s in self.state_freqs])))
        if self.ti_tv and (self.char_model in ['HKY', 'F84']):
            args.append("-t%s" % self.ti_tv)
        if self.general_rates:
            if isinstance(self.general_rates, str):
                args.append("-r%s" % self.general_rates)
            else:
                args.append("-r%s" % (",".join([str(r) for r in self.general_rates])))
        if self.ancestral_seq_idx:
            args.append("-k%s" % self.ancestral_seq_idx)
        if self.write_ancestral_seqs:
            args.append("-wa")
        if self.write_site_rates:
            args.append("-wr")

        if output_format == "nexus":
            args.append("-on")
        elif output_format == "fasta":
            args.append("-of")
        elif output_format == "phylip":
            args.append("-op")
        elif output_format == "relaxed-phylip":
            args.append("-or")
        elif output_format:
            raise ValueError(output_format)

        if append_file:
            args.append("-x%s" % append_file)

        # following are controlled directly by the wrapper
        # silent running
        args.append("-q")
        # we explicitly pass a random number seed on each call
        args.append("-z%s" % self.rng.randint(0, sys.maxsize))
        # force one dataset at a time
        args.append("-n1")
        return args

    def generate(
            self,
            trees,
            dataset=None,
            taxon_namespace=None,
            input_sequences=None,
            tree_serialization_kwargs=None,
            **kwargs
    ):
        stdout = self.generate_raw(
            trees=trees,
            taxon_namespace=taxon_namespace,
            input_sequences=input_sequences,
            tree_serialization_kwargs=tree_serialization_kwargs,
            output_format="nexus",
        )
        if taxon_namespace is None:
            taxon_namespace = trees.taxon_namespace
        if dataset is None:
            dataset = dendropy.DataSet(**kwargs)
            if taxon_namespace is not None:
                dataset.attach_taxon_namespace(taxon_namespace)
        dataset.read(data=stdout, schema="nexus")
        return dataset

    def generate_dicts(
            self,
            **kwargs
    ):
        if "output_format" in kwargs:
            raise TypeError("Cannot specify 'output_format' when requesting 'dict' result")
        kwargs["output_format"] = "fasta"
        if "append_file" in kwargs:
            raise TypeError("Cannot specify 'append_file' when requesting 'dict' result")
        # with self.get_tempfile() as textf:
        with self.get_tempfile() as textf:
            textf.write(self.record_separator)
            textf.write("\n")
            textf.flush()
            kwargs["append_file"] = textf.name
            raw_result = self.generate_raw(**kwargs)
        data_ds = []
        data_d = {}
        label_data = None
        sequence_data = []
        for idx, line in enumerate(raw_result.split("\n")):
            line = line.strip()
            if not line:
                continue
            if line.startswith(">") or line == self.record_separator:
                if label_data is not None:
                    assert label_data not in data_d
                    data_d[label_data] = "".join(sequence_data)
                    sequence_data = []
                else:
                    assert not sequence_data
                if line == self.record_separator:
                    if data_d:
                        data_ds.append(data_d)
                        data_d = {}
                        label_data = None
                else:
                    label_data = line[1:]
            else:
                sequence_data.append(line)
        return data_ds

    def generate_raw(
            self,
            trees,
            taxon_namespace=None,
            input_sequences=None,
            tree_serialization_kwargs=None,
            **kwargs
    ):
        args=self._compose_arguments(**kwargs)
        # with open("x.txt", "w") as inputf:
        with self.get_tempfile() as inputf:
            if input_sequences is not None:
                input_sequences.write_to_stream(inputf, schema="phylip",)
                inputf.write("{}\n".format(len(trees)))
            if tree_serialization_kwargs is None:
                tree_serialization_kwargs = {}
            trees.write_to_stream(
                    inputf,
                    "newick",
                    suppress_rooting=True,
                    suppress_internal_node_labels=True,
                    **tree_serialization_kwargs
            )
            inputf.flush()
            args.append(inputf.name)
            # print("seq-gen args: = %s" % " ".join(args))
            run = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = processio.communicate(run)
            if stderr or run.returncode != 0:
                raise RuntimeError("Seq-gen error: %s" % stderr)
            return stdout





